from numpy import zeros
from numpy import identity
from numpy import hstack
from numpy import vstack
from numpy import sign
from numpy import divide
from numpy import sum as npsum
from numpy import multiply

from numpy import cross
from numpy.linalg import norm
from numpy import eye
from numpy import inner
from numpy import dstack
from numpy import array

from scipy.sparse.linalg import splu
from scipy.sparse import diags

from compas_tno.algorithms import q_from_variables
from compas_tno.algorithms import xyz_from_q
from compas_tno.algorithms import weights_from_xyz


def d_fobj(fobj, x0, eps, *args):
    """Gradient approximated by hand using finite differences.

    Parameters
    ----------
    fconstr : callable
        Function with the constraints
    x0 : array
        Point to compute the pertubation
    eps : float
        Size of the pertubations

    Returns
    -------
    df0dx : array
        Gradient of the function computed using finite diferences
    """

    f0val = fobj(x0, *args)
    n = len(x0)
    df0dx = zeros((n, 1))
    for i in range(n):
        diff = zeros((n, 1))
        diff[i] = eps
        df0dx[i] = (fobj(x0 + diff, *args) - f0val)/diff[i]

    return df0dx


def compute_dQ(q, ind, dep, Edinv, Ei):
    """Sensitivity of (all) the force densities with regards to the independent force densities.

    Parameters
    ----------
    q : array
        Force densities
    ind : list
        Indices of the independent edges
    dep : list
        Indices of the dependent edges
    Edinv : array
        Equilibrium matrix of the dependents inverted
    Ei : array
        Equilibrum matrix of the independents

    Returns
    -------
    dQ : array
        Sensitivities of all q's with respect to the independents
    dQdep : array
        Sensitivities of dependent q's with respect to the independents
    """

    dQdep = Edinv.dot(Ei)
    dQ = zeros((len(q), len(ind)))
    dQ[ind] = identity(len(ind))
    dQ[dep] = dQdep[:, :len(ind)]

    return dQ, dQdep


def deriv_weights_from_matrices(xyz, F, V0, V1, V2, thk=0.5, density=20.0, features=['fixed']):
    """Derivatives of the tributary weights with respect to the position of the nodes based on the assembled sparse matrices linking the topology

    Parameters
    ----------
    xyz : array [n x 3]
        The XYZ coordinates of the thrust network
    F : array [f x n]
        Linear transformation from ``X`` [n x 3] to ``c`` [f x 3] with the position of the centroids
    V0 : array [g x n]
        Mark the influence of the original point in the calculation
    V1 : array [g x n]
        Mark the influence of the neighbor points the calculation
    V2 : array [g x f]
        Mark the influence of the centroid points the calculation
    thk : float, optional
        The thickness of the structure to compute the tributary volumes, by default 0.5
    density : float, optional
        The density of the structure to compute the applied vertical loads, by default 20.0
    features : list, optional
        The features assigned to the optimisation, by default ['fixed']

    Returns
    -------
    dpzdX : array [n x n x 3]
        The sensitivity of node i's tributary area (1st dimension) with regards to node j's movement in the x, y, z directions (3rd dimension)
    """

    g, n = V0.shape
    v0 = V0.dot(xyz)
    v1 = V1.dot(xyz) - v0
    v2 = V2.dot(F).dot(xyz) - v0

    cp = cross(v1, v2)  # cross products
    ncp = norm(cp, axis=1)  # norm cross products

    K1 = V1 - V0
    K2 = V2.dot(F) - V0

    Idt = eye(3)

    dnormdX = zeros((g, n, 3))

    if 'fixed' in features:
        k_range = [2]
    else:
        k_range = [0, 1, 2]

    for k in k_range:
        for i in range(g):
            for j in range(n):
                if K1[i][j] == 0 and K2[i][j] == 0:  # K1 and K2 are extremely sparse
                    continue                         # check scipy.sparse.csr_matrix.nonzero
                dv1i = K1[i][j] * Idt[k]
                dv2i = K2[i][j] * Idt[k]
                dcpi = cross(dv1i, v2[i]) + cross(v1[i], dv2i)
                dnormdX[i][j][k] = inner(cp[i], dcpi)/ncp[i]

    dareadx = V0.transpose().dot(dnormdX[:, :, 0])
    daready = V0.transpose().dot(dnormdX[:, :, 1])
    dareadz = V0.transpose().dot(dnormdX[:, :, 2])

    dpzdX = 0.25 * thk * density * dstack((dareadx, daready, dareadz))

    return dpzdX


def gradient_feasibility(variables, M):
    """Sensitivity of the feasibility objective function, which returns a null vector.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """
    return zeros((len(variables), 1))


def gradient_reduce_thk(variables, M):
    """Sensitivity of the objective function to minimise the thickness.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """
    grad = zeros((len(variables), 1))
    grad[-1] = 1.0
    return grad


def gradient_tight_crosssection(variables, M):
    """Sensitivity of the objective function to tight the cross section.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """
    grad = zeros((len(variables), 1))
    grad[-1] = -1.0
    return grad


def gradient_fmin(variables, M):
    """Sensitivity of the objective function to minimise the thrust.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """

    if isinstance(M, list):
        M = M[0]

    n = M.n
    k = M.k
    is_xyb_var = False
    is_zb_var = False
    update_geometry = False

    k = M.k
    nb = len(M.fixed)

    P_Xh_fixed = M.P[M.fixed][:, :2]  # Horizontal loads in the fixed vertices
    P_free = M.P[M.free]  # Loads in the free vertices

    X = M.X
    U = diags(M.C @ X[:, 0])  # U = diag(Cx)
    V = diags(M.C @ X[:, 1])  # V = diag(Cy)

    dxdq = zeros((n, k))
    dydq = zeros((n, k))

    SPLU_D = None

    qid = variables[:k].reshape(-1, 1)
    q = q_from_variables(qid, M.B, M.d)
    Q = diags(q.flatten())

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
        is_xyb_var = True
        update_geometry = True
    if 'zb' in M.variables:
        zb = variables[-nb:]
        X[M.fixed, 2] = zb.flatten()
        is_zb_var = True
        if 'fixed' not in M.features:
            update_geometry = True

    if update_geometry:
        if 'update-loads' in M.features:
            CitQCi = M.Cit @ Q @ M.Ci
            SPLU_D = splu(CitQCi)
            X[M.free] = xyz_from_q(q, P_free, X[M.fixed], M.Ci, M.Cit, M.Cb, SPLU_D=SPLU_D)
            pz = -1 * weights_from_xyz(X, M.F, M.V0, M.V1, M.V2, thk=M.thk, density=M.ro)
            P_free[:, 2] = pz[M.free]

        X[M.free] = xyz_from_q(q, P_free, X[M.fixed], M.Ci, M.Cit, M.Cb, SPLU_D=SPLU_D)
        U = diags(M.C @ X[:, 0])  # U = diag(Cx)
        V = diags(M.C @ X[:, 1])  # V = diag(Cy)

        if not SPLU_D:
            CitQCi = M.Cit @ Q @ M.Ci
            SPLU_D = splu(CitQCi)

        dxidq = SPLU_D.solve((-M.Cit.dot(U)).toarray()).dot(M.B)
        dxdq = zeros((n, k))
        dxdq[M.free] = dxidq

        dyidq = SPLU_D.solve((-M.Cit.dot(V)).toarray()).dot(M.B)
        dydq = zeros((n, k))
        dydq[M.free] = dyidq

    CfU = M.Cb.transpose() @ U
    CfV = M.Cb.transpose() @ V
    dRxdq = CfU @ M.B + M.Cb.transpose() @ Q @ M.C @ dxdq
    dRydq = CfV @ M.B + M.Cb.transpose() @ Q @ M.C @ dydq

    Rx = (CfU @ q - P_Xh_fixed[:, [0]])  # check this +/- business
    Ry = (CfV @ q - P_Xh_fixed[:, [1]])
    R = norm(hstack([Rx, Ry]), axis=1).reshape(-1, 1)

    Rx_over_R = divide(Rx, R)
    Ry_over_R = divide(Ry, R)

    gradient = (Rx_over_R.transpose() @ dRxdq + Ry_over_R.transpose() @ dRydq).transpose()

    if is_xyb_var:
        dxdxb = zeros((n, nb))
        CitQCf = (M.Cit @ Q @ M.Cb).toarray()
        dxdxb[M.free] = SPLU_D.solve(-CitQCf)
        dxdxb[M.fixed] = identity(nb)
        dydyb = dxdxb
        dRxdx = M.Cb.transpose() @ Q @ M.C @ dxdxb
        dRydy = M.Cb.transpose() @ Q @ M.C @ dydyb
        gradient_xb = (Rx_over_R.transpose() @ dRxdx).transpose()
        gradient_yb = (Ry_over_R.transpose() @ dRydy).transpose()
        gradient = vstack([gradient, gradient_xb, gradient_yb])
    if is_zb_var:
        gradient = vstack([gradient, zeros((nb, 1))])
    if 'delta' in M.variables:
        gradient = vstack([gradient, zeros((1, 1))])

    return array(gradient).flatten()


def gradient_fmax(variables, M):
    """Sensitivity of the objective function to maximise the thrust.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """

    return -1 * gradient_fmin(variables, M)


def gradient_bestfit(variables, M):
    """Sensitivity of the objective function to minimise the vertical squared distance to the target.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """

    if isinstance(M, list):
        M = M[0]

    n = M.n
    k = M.k

    k = M.k
    nb = len(M.fixed)
    n = M.n

    qid = variables[:k].reshape(-1, 1)
    M.q = q_from_variables(qid, M.B, M.d)

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = variables[-nb:]
        M.X[M.fixed, [2]] = zb.flatten()

    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    M.W = diags(M.C.dot(M.X[:, 2]))  # W = diag(Cz)

    f = 2*(M.X[:, [2]] - M.s)

    Q = diags(M.q.ravel())
    CitQCi = M.Cit.dot(Q).dot(M.Ci)
    SPLU_D = splu(CitQCi)

    dzdq = zeros((n, k))
    dzidq = SPLU_D.solve(-M.Cit.dot(M.W).toarray()).dot(M.B)
    dzdq[M.free] = dzidq

    gradient = (f.transpose().dot(dzdq)).transpose()

    if 'zb' in M.variables:
        dz_dzb = zeros((n, nb))
        CitQCf = M.Cit.dot(Q).dot(M.Cb).toarray()
        dz_dzb[M.free] = SPLU_D.solve(-CitQCf)
        dz_dzb[M.fixed] = identity(nb)

        gradient_zb = (f.transpose().dot(dz_dzb)).transpose()
        gradient = vstack([gradient, gradient_zb])

    return gradient


def gradient_horprojection(variables, M):
    """Sensitivity of the objective function to minimise the horizontal squared distance of the nodes on the form diagram to a given pattern.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """

    if isinstance(M, list):
        M = M[0]

    n = M.n
    k = M.k

    k = M.k
    nb = len(M.fixed)
    n = M.n

    qid = variables[:k].reshape(-1, 1)
    M.q = q_from_variables(qid, M.B, M.d)

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = variables[-nb:]
        M.X[M.fixed, [2]] = zb.flatten()

    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    M.U = diags(M.C.dot(M.X[:, 0]))  # U = diag(Cx)
    M.V = diags(M.C.dot(M.X[:, 1]))  # V = diag(Cy)
    # M.W = diags(M.C.dot(M.X[:, 2]))  # W = diag(Cz)

    fx = 2*(M.X[:, [0]] - M.x0)
    fy = 2*(M.X[:, [1]] - M.y0)

    Q = diags(M.q.ravel())
    CitQCi = M.Cit.dot(Q).dot(M.Ci)
    SPLU_D = splu(CitQCi)

    dxdq = zeros((n, k))
    dxidq = SPLU_D.solve(-M.Cit.dot(M.U).toarray()).dot(M.B)
    dxdq[M.free] = dxidq

    dydq = zeros((n, k))
    dyidq = SPLU_D.solve(-M.Cit.dot(M.V).toarray()).dot(M.B)
    dydq[M.free] = dyidq

    # dzdq = zeros((n, k))
    # dzidq = SPLU_D.solve(-M.Cit.dot(M.W).toarray()).dot(M.B)
    # dzdq[M.free] = dzidq

    gradient_x = (fx.transpose().dot(dxdq)).transpose()
    gradient_y = (fy.transpose().dot(dydq)).transpose()

    gradient = gradient_x + gradient_y

    if 'zb' in M.variables:
        gradient_zb = zeros((nb, 1))
        gradient = vstack([gradient, gradient_zb])

    return gradient


def gradient_complementary_energy(variables, M):
    """Sensitivity of the objective function to minimise the complementary energy.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """

    if isinstance(M, list):
        M = M[0]

    n = M.n
    k = M.k
    is_xyb_var = False
    is_zb_var = False

    k = M.k
    nb = len(M.fixed)

    qid = variables[:k].reshape(-1, 1)
    M.q = q_from_variables(qid, M.B, M.d)

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
        is_xyb_var = True
    if 'zb' in M.variables:
        zb = variables[-nb:]
        M.X[M.fixed, [2]] = zb.flatten()
        is_zb_var = True

    # update geometry
    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    M.U = diags(M.C.dot(M.X[:, 0]))  # U = diag(Cx)
    M.V = diags(M.C.dot(M.X[:, 1]))  # V = diag(Cy)
    M.W = diags(M.C.dot(M.X[:, 2]))  # W = diag(Cz)

    Q = diags(M.q.ravel())
    CitQCi = M.Cit.dot(Q).dot(M.Ci)
    SPLU_D = splu(CitQCi)

    dxidq = SPLU_D.solve((-M.Cit.dot(M.U)).toarray()).dot(M.B)
    dxdq = zeros((n, k))
    dxdq[M.free] = dxidq

    dyidq = SPLU_D.solve((-M.Cit.dot(M.V)).toarray()).dot(M.B)
    dydq = zeros((n, k))
    dydq[M.free] = dyidq

    dzidq = SPLU_D.solve(-M.Cit.dot(M.W).toarray()).dot(M.B)
    dzdq = zeros((n, k))
    dzdq[M.free] = dzidq

    CfU = M.Cb.transpose().dot(M.U)
    CfV = M.Cb.transpose().dot(M.V)
    CfW = M.Cb.transpose().dot(M.W)
    dRxdq = CfU.dot(M.B) + M.Cb.transpose().dot(Q).dot(M.C).dot(dxdq)
    dRydq = CfV.dot(M.B) + M.Cb.transpose().dot(Q).dot(M.C).dot(dydq)
    dRzdq = CfW.dot(M.B) + M.Cb.transpose().dot(Q).dot(M.C).dot(dzdq)

    gradient = (M.dXb[:, [0]].transpose().dot(dRxdq) + M.dXb[:, [1]].transpose().dot(dRydq) + M.dXb[:, [2]].transpose().dot(dRzdq)).transpose()

    CitQCf = M.Cit.dot(Q).dot(M.Cb).toarray()
    if is_xyb_var:
        dxdxb = zeros((n, nb))
        dxdxb[M.free] = SPLU_D.solve(-CitQCf)
        dxdxb[M.fixed] = identity(nb)
        dydyb = dxdxb
        dRxdxb = M.Cb.transpose().dot(Q).dot(M.C).dot(dxdxb)
        dRydyb = M.Cb.transpose().dot(Q).dot(M.C).dot(dydyb)
        gradient_xb = (M.dXb[:, [0]].transpose().dot(dRxdxb)).transpose()
        gradient_yb = (M.dXb[:, [1]].transpose().dot(dRydyb)).transpose()
        gradient = vstack([gradient, gradient_xb, gradient_yb])
    if is_zb_var:
        dzdzb = zeros((n, nb))
        dzdzb[M.free] = SPLU_D.solve(-CitQCf)
        dzdzb[M.fixed] = identity(nb)
        dRzdzb = M.Cb.transpose().dot(Q).dot(M.C).dot(dzdzb)
        gradient_zb = (M.dXb[:, [2]].transpose().dot(dRzdzb)).transpose()
        gradient = vstack([gradient, gradient_zb])

    return -1 * array(gradient).flatten()


def gradient_complementary_energy_nonlinear(variables, M):
    """Sensitivity of the objective function to minimise nonlinear complementary energy.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """

    grad_lin = gradient_complementary_energy(variables, M)

    if M.Ecomp_method == 'simplified':  # assuming area and lengths constant - computed in beginning

        nb = len(M.fixed)

        dEdq_vector = 2 * M.stiff * M.q.reshape(-1, 1)
        dEdq = diags(dEdq_vector.flatten())
        dEdqid = dEdq.dot(M.B)
        grad_quad = npsum(dEdqid, axis=0).reshape(-1, 1)

        if 'xyb' in M.variables:
            grad_quad = vstack([grad_quad, zeros((2*nb, 1))])
        if 'zb' in M.variables:
            grad_quad = vstack([grad_quad, zeros((nb, 1))])

    elif M.Ecomp_method == 'complete':
        grad_quad = gradient_loadpath(variables, M) * M.stiff

    fgrad = grad_lin + grad_quad.flatten()

    return fgrad


def gradient_loadpath(variables, M):
    """Sensitivity of the objective function to minimise the loadpath.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """

    if isinstance(M, list):
        M = M[0]

    n = M.n
    k = M.k

    k = M.k
    nb = len(M.fixed)
    n = M.n

    qid = variables[:k].reshape(-1, 1)
    M.q = q_from_variables(qid, M.B, M.d)

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = variables[-nb:]
        M.X[M.fixed, [2]] = zb.flatten()

    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    uvw = M.C.dot(M.X)
    l2 = npsum(uvw**2, axis=1).reshape(-1, 1)

    M.U = diags(uvw[:, 0])  # U = diag(Cx)
    M.V = diags(uvw[:, 1])  # V = diag(Cy)
    M.W = diags(uvw[:, 2])  # W = diag(Cz)

    Q = diags(M.q.ravel())
    CitQCi = M.Cit.dot(Q).dot(M.Ci)
    SPLU_D = splu(CitQCi)

    dxdq = zeros((n, k))
    dydq = zeros((n, k))
    dzdq = zeros((n, k))

    dxidq = SPLU_D.solve(-M.Cit.dot(M.U).toarray()).dot(M.B)  # not needed if fixed
    dxdq[M.free] = dxidq

    dyidq = SPLU_D.solve(-M.Cit.dot(M.V).toarray()).dot(M.B)  # not needed if fixed
    dydq[M.free] = dyidq

    dzidq = SPLU_D.solve(-M.Cit.dot(M.W).toarray()).dot(M.B)
    dzdq[M.free] = dzidq

    # print(multiply(sign(M.q), l2).transpose().shape)
    # print(multiply(sign(M.q), l2).transpose().dot(M.B).shape)
    # print(M.U.dot(M.C.dot(dxdq)).shape)
    # print(M.V.dot(M.C.dot(dydq)).shape)
    # print(M.W.dot(M.C.dot(dzdq)).shape)
    # print(M.q.transpose().shape)
    # print(abs(M.q.transpose()) * (M.U.dot(M.C.dot(dxdq)) + M.V.dot(M.C.dot(dydq)) + M.W.dot(M.C.dot(dzdq))).shape)
    # print(M.C.dot(dzdq).shape)

    dudq = M.C.dot(dxdq)
    dvdq = M.C.dot(dydq)
    dwdq = M.C.dot(dzdq)

    dldq_1 = (multiply(sign(M.q).reshape(-1, 1), l2).transpose().dot(M.B)).flatten()

    # gradient = (multiply(sign(M.q), l2).transpose().dot(M.B) + 2*abs(M.q.transpose()).dot(M.U.dot(M.C.dot(dxdq)) + M.V.dot(M.C.dot(dydq)) + M.W.dot(M.C.dot(dzdq)))).transpose()
    gradient = zeros((k, 1))

    for i in range(k):
        dlidqi = multiply(uvw[:, [0]], dudq[:, [i]]) + multiply(uvw[:, [1]], dvdq[:, [i]]) + multiply(uvw[:, [2]], dwdq[:, [i]])
        # gradient[i] = sign(M.q[i]) * l2[i] + 2 * abs(M.q.transpose()).dot(dlidqi)
        gradient[i] = dldq_1[i] + 2*abs(M.q.transpose()).dot(dlidqi)

    if 'zb' in M.variables:
        dz_dzb = zeros((n, nb))
        CitQCf = M.Cit.dot(Q).dot(M.Cb).toarray()
        dz_dzb[M.free] = SPLU_D.solve(-CitQCf)
        dz_dzb[M.fixed] = identity(nb)

        gradient_zb = (2*abs(M.q.transpose()).dot(M.W.dot(M.C.dot(dz_dzb)))).transpose().reshape(-1, 1)
        gradient = vstack([gradient, gradient_zb])

    return gradient


def gradient_max_section(variables, M):
    """Sensitivity of the objective function to minimise additional thickness required to find a feasible thrust network.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    grad : array (k x 1)
        The gradient of the objective function in the point. Represents the sensitivity of each variable in the objective function value.
    """

    k = M.k
    nb = len(M.fixed)
    n = M.X.shape[0]

    check = k

    gradient = zeros((k, 1))

    if 'xyb' in M.variables:
        check = check + 2*nb
        gradient = vstack([gradient, zeros((2*nb, 1))])
    if 'zb' in M.variables:
        check = check + nb
        gradient = vstack([gradient, zeros((nb, 1))])
    if 'tub' in M.variables:
        tub = variables[check: check + n].reshape(-1, 1)
        gradient = vstack([gradient, 2 * tub])
        check = check + n
    if 'tlb' in M.variables:
        tlb = variables[check: check + n].reshape(-1, 1)
        gradient = vstack([gradient, 2 * tlb])
        check = check + n
    if 'tub_reac' in M.variables:
        tub_reac = variables[check: check + 2*nb].reshape(-1, 1)
        gradient = vstack([gradient, 2 * tub_reac])  # check if this changes because of the modulus
        check = check + 2*nb

    return gradient



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import errno
import shelve

import scriptcontext as sc

import compas
import compas_rhino

from compas.rpc import Proxy
from compas_tno.rhino import Scene
# from compas_tno.rhino import Browser  # From compas_ags, check if needed to do forms
from compas_tno.activate import check
from compas_tno.activate import activate


__commandname__ = "TNO__init"


# the current working directory could be the APPTEMP directory for AGS
# until it is specified by the user
HERE = compas_rhino.get_document_dirname()
HOME = os.path.expanduser('~')
CWD = HERE or HOME


SETTINGS = {
    'TNO': {
        'autoupdate': True,
    }
}


def RunCommand(is_interactive):

    if check():
        print("Current plugin is already activated")
    else:
        compas_rhino.rs.MessageBox("Detected environment change, re-activating plugin", 0, "Re-activating Needed")
        if activate():
            compas_rhino.rs.MessageBox("Restart Rhino for the change to take effect", 0, "Restart Rhino")
        else:
            compas_rhino.rs.MessageBox("Someting wrong during re-activation", 0, "Error")
        return

    shelvepath = os.path.join(compas.APPTEMP, 'AGS', '.history')
    if not os.path.exists(os.path.dirname(shelvepath)):
        try:
            os.makedirs(os.path.dirname(shelvepath))
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise

    db = shelve.open(shelvepath, 'n')
    db['states'] = []

    scene = Scene(db, 20, SETTINGS)
    scene.purge()

    sc.sticky["TNO"] = {
        'proxy': Proxy(),
        'system': {
            "session.dirname": CWD,
            "session.filename": None,
            "session.extension": 'tno'
        },
        'scene': scene,
    }

    scene.update()

    # would be useful to add a notification about the cloud: new / reconnect
    # compas_rhino.display_message("AGS has started.")
    # Browser()  # working on a Browser

    # code to install in python:
    # python -m compas_tno.install


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from compas_tno.problems import initialise_problem_general

from compas_tno.problems import adapt_problem_to_fixed_diagram
from compas_tno.problems import adapt_problem_to_sym_diagram
from compas_tno.problems import adapt_problem_to_sym_and_fixed_diagram

from compas_tno.problems import objective_selector

from compas_tno.problems import callback_save_json
from compas_tno.problems import callback_create_json

from compas_tno.problems import initialize_loadpath
from compas_tno.problems import initialize_tna

from compas_tno.algorithms import apply_sag
from compas_tno.algorithms import equilibrium_fdm
from compas_tno.algorithms import q_from_variables
from compas_tno.algorithms import xyz_from_q

from compas_tno.problems import sensitivities_wrapper

from compas_tno.problems import constr_wrapper

from compas_tno.plotters import TNOPlotter

from compas_tno.utilities import apply_bounds_on_q
from compas_tno.utilities import compute_form_initial_lengths
from compas_tno.utilities import compute_edge_stiffness
from compas_tno.utilities import set_b_constraint

from numpy import append
from numpy import array
from numpy import zeros
from numpy import vstack
from numpy import diag

from math import isnan


def set_up_general_optimisation(analysis):
    """ Set up a nonlinear optimisation problem.

    Parameters
    ----------
    analysis : :class:`~compas_tno.analysis.Analysis`
        Analysis object with information about optimiser, form and shape.

    Returns
    -------
    analysis : :class:`~compas_tno.analysis.Analysis`
        Analysis object set up for optimise.

    """

    form = analysis.form
    optimiser = analysis.optimiser
    shape = analysis.shape

    printout = optimiser.settings.get('printout', True)
    plot = optimiser.settings.get('plot', False)
    axis_symmetry = optimiser.settings.get('axis_sym', None)
    sym_loads = optimiser.settings.get('sym_loads', False)
    fjac = optimiser.settings.get('jacobian', False)
    starting_point = optimiser.settings.get('starting_point', 'current')
    find_inds = optimiser.settings.get('find_inds', False)
    tol_inds = optimiser.settings.get('tol_inds', None)
    method_ind = optimiser.settings.get('method_ind', 'SVD')
    qmin = optimiser.settings.get('qmin', -1e+4)
    qmax = optimiser.settings.get('qmax', +1e-8)
    features = optimiser.settings.get('features', [])
    save_iterations = optimiser.settings.get('save_iterations', False)
    solver_convex = optimiser.settings.get('solver_convex', 'MATLAB')
    autodiff = optimiser.settings.get('autodiff', False)

    pattern_center = form.parameters.get('center', None)

    if shape:
        thk = shape.datashape.get('thk', None)
        form.attributes['thk'] = thk  # for safety storing the thk. If optimisation is minthk, this will be overwritten
    else:
        thk = 0.50

    objective = optimiser.settings['objective']
    variables = optimiser.settings['variables']
    constraints = optimiser.settings['constraints']

    i_k = form.index_key()

    qmin_applied = form.edges_attribute('qmin')
    for qmin_applied_i in qmin_applied:
        if qmin_applied_i is None:
            print('Appied qmin / qmax:', qmin, qmax)
            apply_bounds_on_q(form, qmin=qmin, qmax=qmax)
            break

    M = optimiser.M
    if not M:
        M = initialise_problem_general(form)

    M.variables = variables
    M.constraints = constraints
    M.features = features
    M.shape = shape
    M.thk = thk

    if 'update-loads' in features:
        F, V0, V1, V2 = form.tributary_matrices(sparse=False)
    else:
        F, V0, V1, V2 = 4*[None]

    M.F = F
    M.V0 = V0
    M.V1 = V1
    M.V2 = V2
    M.ro = shape.ro

    if starting_point == 'current':
        pass
    elif starting_point == 'sag':
        apply_sag(form, boundary_force=50.0)  # the issue here is that after the sag the M.x0, M.y0 are not updated
        initialize_tna(form)
    elif starting_point == 'loadpath':
        printout_loadpath = False  # this need to be a proper verbose level
        initialize_loadpath(form, problem=M, find_inds=find_inds, solver_convex=solver_convex, printout=printout_loadpath)
    elif starting_point == 'relax':
        equilibrium_fdm(form)
        initialize_tna(form)
    elif starting_point == 'tna' or starting_point == 'TNA':
        initialize_tna(form)
    else:
        print('Warning: define starting point')

    M.q = array([form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True})]).reshape(-1, 1)

    if 'fixed' in features and 'sym' in features:
        # print('\n-------- Initialisation with fixed and sym form --------')
        adapt_problem_to_sym_and_fixed_diagram(M, form, method=method_ind, list_axis_symmetry=axis_symmetry,
                                               center=pattern_center, correct_loads=sym_loads, printout=printout, tol=tol_inds)
    elif 'sym' in features:
        # print('\n-------- Initialisation with sym form --------')
        adapt_problem_to_sym_diagram(M, form, list_axis_symmetry=axis_symmetry, center=pattern_center, correct_loads=sym_loads, printout=printout)
    elif 'fixed' in features:
        # print('\n-------- Initialisation with fixed form --------')
        adapt_problem_to_fixed_diagram(M, form, method=method_ind, printout=printout, tol=tol_inds)
    else:
        # print('\n-------- Initialisation with no-fixed and no-sym form --------')
        pass

    # If objective is the complementary energy

    if 'Ecomp' in objective.split('-'):
        M.dXb = optimiser.settings['support_displacement']

    if objective == 'Ecomp-nonlinear':
        Ecomp_method = optimiser.settings.get('Ecomp_method', 'simplified')
        M.Ecomp_method = Ecomp_method

        if Ecomp_method == 'simplified':
            stiff = zeros((M.m, 1))
            lengths = compute_form_initial_lengths(form)
            k = compute_edge_stiffness(form, lengths=lengths)
            for index, edge in enumerate(form.edges()):
                stiff[index] = 1 / 2 * 1 / k[index] * lengths[index] ** 2
            M.stiff = stiff
        elif Ecomp_method == 'complete':
            raise NotImplementedError()

    # Set specific constraints

    if 'reac_bounds' in constraints:
        M.b = set_b_constraint(form, printout)
    else:
        M.b = None

    # Select Objetive, Gradient, Costraints and Jacobian.

    fobj, fgrad = objective_selector(objective)

    fconstr = constr_wrapper
    if fjac:
        fjac = sensitivities_wrapper

    # Alternative for autodiff

    if autodiff:
        raise NotImplementedError('Autodifferentiation is currently not available')

    # Select starting point (x0) and max/min for variables

    x0 = M.q[M.ind]
    bounds = [[qmin.item(), qmax.item()] for i, (qmin, qmax) in enumerate(zip(M.qmin, M.qmax)) if i in M.ind]

    qid = M.q[M.ind]
    M.q = q_from_variables(qid, M.B, M.d)
    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    error = sum((M.E.dot(M.q) - M.ph)**2)
    if error > 0.001:
        print('Warning: Error equilibrium:', error)

    # bounds = [[qmin, qmax]] * M.k

    if 'xyb' in variables:
        xyb0 = M.X[M.fixed, :2].flatten('F').reshape((-1, 1))
        x0 = append(x0, xyb0).reshape(-1, 1)
        bounds_x = []
        bounds_y = []
        for i in M.fixed:
            bounds_x.append([form.vertex_attribute(i_k[i], 'xmin'), form.vertex_attribute(i_k[i], 'xmax')])
            bounds_y.append([form.vertex_attribute(i_k[i], 'ymin'), form.vertex_attribute(i_k[i], 'ymax')])
        bounds = bounds + bounds_x + bounds_y

    if 'zb' in variables:
        zb0 = M.X[M.fixed, 2].flatten('F').reshape((-1, 1))
        x0 = append(x0, zb0).reshape(-1, 1)
        bounds_z = []
        for i in M.fixed:
            bounds_z.append([form.vertex_attribute(i_k[i], 'lb'), form.vertex_attribute(i_k[i], 'ub')])
        bounds = bounds + bounds_z

    if 't' in variables:
        min_thk = optimiser.settings.get('min_thk', 0.001)
        max_thk = optimiser.settings.get('max_thk', thk)
        x0 = append(x0, thk).reshape(-1, 1)
        bounds = bounds + [[min_thk, max_thk]]

    if 'n' in variables:
        thk0_approx = thk  # shape.datashape['thk']
        print('Thickness approximate:', thk0_approx)
        x0 = append(x0, 0.0).reshape(-1, 1)
        min_limit = 0.0  # /2  # 0.0
        bounds = bounds + [[min_limit, thk0_approx/2]]
        # bounds = bounds + [[min_limit, thk0_approx/2]]

    if 'tub' in variables:
        M.tub = zeros((M.n, 1))
        tubmax = form.vertices_attribute('tubmax')
        M.tubmax = array(tubmax).reshape(M.n, 1)
        M.tubmin = zeros((M.n, 1))
        x0 = append(x0, M.tub)
        bounds = bounds + list(zip(M.tubmin, M.tubmax))

    if 'tlb' in variables:
        M.tlb = zeros((M.n, 1))
        tlbmax = form.vertices_attribute('tlbmax')
        M.tlbmax = array(tlbmax).reshape(M.n, 1)
        M.tlbmin = zeros((M.n, 1))
        x0 = append(x0, M.tlb)
        bounds = bounds + list(zip(M.tlbmin, M.tlbmax))

    if 'tub_reac' in variables:
        tub_reac = []
        for key in form.vertices_where({'is_fixed': True}):
            tub_reac.append(form.vertex_attribute(key, 'tub_reacmax'))
        tub_reac = array(tub_reac)
        tub_reac = vstack([tub_reac[:, 0].reshape(-1, 1), tub_reac[:, 1].reshape(-1, 1)])
        M.tub_reac = abs(tub_reac)
        M.tub_reac_min = zeros((2*M.nb, 1))
        x0 = append(x0, M.tub_reac_min)
        bounds = bounds + list(zip(M.tub_reac_min, M.tub_reac))

    if 'lambdh' in variables:
        lambd0 = 1.0
        direction = optimiser.settings.get('load_direction', None).reshape(-1, 1)
        M.px0 = direction[:M.n].reshape(-1, 1)
        M.py0 = direction[M.n: 2*M.n].reshape(-1, 1)
        max_lambd = optimiser.settings.get('max_lambd', lambd0 * 10)
        min_lambd = 0.0
        x0 = append(x0, lambd0).reshape(-1, 1)
        bounds = bounds + [[min_lambd, max_lambd]]

    if 'lambdv' in variables:
        direction = array(optimiser.settings.get('load_direction', None)).reshape(-1, 1)
        max_lambd = optimiser.settings.get('max_lambd', 100.0)
        min_lambd = 0.0
        lambd0 = 1.0
        M.pzv = direction
        M.pz0 = array(form.vertices_attribute('pz')).reshape(-1, 1)
        x0 = append(x0, lambd0).reshape(-1, 1)
        bounds = bounds + [[min_lambd, max_lambd]]

    if 'delta' in variables:
        dX = optimiser.settings['displ_map']
        max_delta = optimiser.settings.get('max_delta', 1.0)
        min_delta = 0.0
        delta0 = optimiser.settings.get('delta0', 0.0)
        M.dX = dX
        x0 = append(x0, delta0).reshape(-1, 1)
        bounds = bounds + [[min_delta, max_delta]]

        Ud = diag(M.C @ dX[:, 0])
        Vd = diag(M.C @ dX[:, 1])
        Edx = M.Cit @ Ud
        Edy = M.Cit @ Vd
        M.Ed = vstack([Edx, Edy])

    if save_iterations:
        callback_create_json()
        optimiser.callback = callback_save_json
        callback_save_json(x0)  # save staring point to file

    if plot:
        plotter = TNOPlotter(form)
        plotter.draw_form_independents()
        plotter.show()
        if 'sym' in features:
            plotter = TNOPlotter(form)
            plotter.draw_form_sym(print_sym=True)
            plotter.show()

    if any([isnan(M.ub[i]) for i in range(len(M.ub))]) or any([isnan(M.lb[i]) for i in range(len(M.lb))]):
        print('Is Nan for the bounds. Optimisation can not proceed')
        raise ValueError('Check bounds that constraint nodes')

    f0 = fobj(x0, M)
    g0 = fconstr(x0, M)

    if fgrad:
        grad = fgrad(x0, M)
    if fjac:
        jac = fjac(x0, M)

    for i, key in enumerate(form.vertices()):
        form.vertex_attribute(key, 'x', M.X[i, 0])
        form.vertex_attribute(key, 'y', M.X[i, 1])
        form.vertex_attribute(key, 'z', M.X[i, 2])

    if plot:
        from compas_tno.viewers import Viewer
        view = Viewer(form)
        view.draw_thrust()
        view.show()

    if printout:
        print('-'*20)
        print('NPL (Non Linear Problem) Data:')
        print('Number of force variables:', len(M.ind))
        print('Number of variables:', len(x0))
        print('Number of constraints:', len(g0))
        if 'funicular' in M.constraints:
            print('# constraints funicular:', 2*len(M.q))
        if 'envelopexy' in M.constraints:
            print('# constraints envelope xy:', 4*len(M.X))
        if 'envelope' in M.constraints:
            print('# constraints envelope z:', 2*len(M.X))
        if 'reac_bounds' in M.constraints:
            print('# constraints reac_bounds:', 2*len(M.fixed))
        if fgrad:
            print('Shape of gradient:', grad.shape)
        if fjac:
            print('Shape of jacobian:', jac.shape)
        print('Init. Objective Value: {0}'.format(f0))
        print('Init. Constraints Extremes: {0:.3f} to {1:.3f}'.format(max(g0), min(g0)))
        violated = []
        for i in range(len(g0)):
            if g0[i] < 0:
                violated.append(i)
        if violated:
            print('Constraints Violated #:', violated)

    optimiser.fobj = fobj
    optimiser.fconstr = fconstr
    optimiser.fgrad = fgrad
    optimiser.fjac = fjac
    optimiser.M = M
    optimiser.x0 = x0
    optimiser.bounds = bounds
    optimiser.f0 = f0
    optimiser.g0 = g0

    analysis.form = form
    analysis.optimiser = optimiser

    return analysis


def set_up_convex_optimisation(analysis):
    """ Set up a convex optimisation problem.

    Parameters
    ----------
    analysis : :class:`~compas_tno.analysis.Analysis`
        Analysis object with information about optimiser, form and shape.

    Returns
    -------
    analysis : :class:`~compas_tno.analysis.Analysis`
        Analysis object set up for optimise.

    """

    form = analysis.form
    optimiser = analysis.optimiser
    qmax = optimiser.settings['qmax']
    qmin = optimiser.settings['qmin']

    objective = optimiser.settings['objective']
    variables = optimiser.settings['variables']
    constraints = optimiser.settings['constraints']

    # Select Objective

    if objective not in ['loadpath', 'feasibility']:
        print('Warning: Non-convex problem for the objective: ', objective, '. Try changing the objective to \'loadpath\' or \'fesibility\'.')

    if variables == ['q']:
        pass
    else:
        print('Warning: Non-convex problem for the variables: ', variables, '. Considering only \'q\' instead and assuming coplanar supports (zb=0).')

    if constraints == ['funicular']:
        pass
    else:
        print('Warning: Non-convex problem for the constraints: ', constraints, '. Considering only \'funicular\' instead.')

    apply_bounds_on_q(form, qmin=qmin, qmax=qmax)

    M = initialise_problem_general(form)
    M.variables = variables
    M.constraints = constraints

    optimiser.M = M

    analysis.form = form
    analysis.optimiser = optimiser

    return analysis



from compas.datastructures import Mesh
from compas.utilities import geometric_key
import math


def create_arch_form_diagram(cls, H=1.0, L=2.0, x0=0.0, discretisation=100):
    """Construct a FormDiagram based on an arch linear discretisation.
    Note: The nodes of the form diagram are spaced following a projection in a semicircular arch.

    Parameters
    ----------
    H : float, optional
        Height of the arch, by default 1.00
    L : float, optional
        Span of the arch, by default 2.00
    x0 : float, optional
        Initial coordiante of the arch, by default 0.0
    discretisation : int, optional
        Numbers of nodes to be considered in the form diagram, by default 100

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.
    """

    # Add option for starting from Hi and Li for a given thk.

    radius = H / 2 + (L**2 / (8 * H))
    print('radius =', radius)
    spr = math.atan2((L/2), (radius - H))
    print('springing angle =', math.degrees(spr))
    tot_angle = 2*spr
    angle_init = (math.pi - tot_angle)/2
    an = tot_angle / (discretisation - 1)
    lines = []
    gkey_fix = []

    for i in range(discretisation-1):
        angle_i = angle_init + i * an
        angle_f = angle_init + (i + 1) * an
        xi = L/2 - radius * math.cos(angle_i) + x0
        xf = L/2 - radius * math.cos(angle_f) + x0
        lines.append([[xi, 0.0, 0.0], [xf, 0.0, 0.0]])
        if i == 0:
            gkey_fix.append(geometric_key([xi, 0.0, 0.0], precision=6))
        elif i == discretisation - 2:
            gkey_fix.append(geometric_key([xf, 0.0, 0.0], precision=6))

    mesh = Mesh.from_lines(lines)
    form = cls.from_mesh(mesh)
    gkey_key = form.gkey_key(precision=6)

    form.vertex_attribute(gkey_key[gkey_fix[0]], 'is_fixed', True)
    form.vertex_attribute(gkey_key[gkey_fix[1]], 'is_fixed', True)

    return form


def create_linear_form_diagram(cls, L=2.0, x0=0.0, discretisation=100):
    """ Helper to create a arch linear form-diagram with equaly spaced (in 2D) nodes.

    Parameters
    ----------
    L : float, optional
        Span of the arch, by default 2.00
    x0 : float, optional
        Initial coordiante of the arch, by default 0.0
    discretisation : int, optional
        Numbers of nodes to be considered in the form diagram, by default 100

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        FormDiagram generated according to the parameters.

    """

    from numpy import linspace
    x = linspace(x0, x0 + L, discretisation)  # Continue this remove need of numpy in the future
    lines = []
    gkey_fix = []

    for i in range(discretisation-1):
        xi = x[i]
        xf = x[i + 1]
        lines.append([[xi, 0.0, 0.0], [xf, 0.0, 0.0]])
        if i == 0:
            gkey_fix.append(geometric_key([xi, 0.0, 0.0], precision=6))
        elif i == discretisation - 2:
            gkey_fix.append(geometric_key([xf, 0.0, 0.0], precision=6))

    mesh = Mesh.from_lines(lines)
    form = cls.from_mesh(mesh)
    gkey_key = form.gkey_key(precision=6)

    form.vertex_attribute(gkey_key[gkey_fix[0]], 'is_fixed', True)
    form.vertex_attribute(gkey_key[gkey_fix[1]], 'is_fixed', True)

    return form



from compas_tno.algorithms import xyz_from_q

try:
    import matlab.engine
except BaseException:
    pass
    # print('matlab not available in the system')
import time

from numpy import array

from compas_tno.algorithms import compute_reactions

from compas_tno.problems import initialise_problem_general
from compas_tno.problems import adapt_problem_to_fixed_diagram


def run_optimisation_MATLAB(analysis):
    """ Run convex optimisation problem with MATLAB after going through the optimisation set up.

    Parameters
    ----------
    obj : :class:`~compas_tno.analysis.Analysis`
        Analysis object with information about optimiser, form and shape.

    Returns
    -------
    obj : dict
        Dictionary with the returned values from the convex optimisation.

    """

    form = analysis.form
    problem = analysis.optimiser.M
    find_inds = analysis.optimiser.settings.get('find_inds', False)
    printout = analysis.optimiser.settings.get('printout', False)

    problem = run_loadpath_from_form_MATLAB(form, problem=problem, find_inds=find_inds, printout=printout)

    return problem


def run_loadpath_from_form_MATLAB(form, problem=None, find_inds=False, printout=False):
    """ Run convex optimisation problem with MATLAB directly from the Form Diagram
        OBS: Requires installation of CVX and MATLAB.

    Parameters
    ----------
    form : FormDiagram
        FormDiagram object with form containing full information.
    problem : Problem, optional
        The problem with matrices of interest, by default None
    find_inds : bool, optional
        Whether or not independents must be computed before the analysis, by default False
    printout : bool, optional
        Whether or not print results, by default False

    Returns
    -------
    obj : dict
        Dictionary with the returned values from the convex optimisation.
    """

    # try:
    #     import matlab.engine
    # except BaseException:
    #     print('Warning: Matlab not available in the system')

    # future = matlab.engine.connect_matlab(background=True)
    future = matlab.engine.start_matlab(background=True)

    eng = future.result()

    if not problem:
        problem = initialise_problem_general(form)

    if find_inds:
        adapt_problem_to_fixed_diagram(problem, form)

    problem = call_and_output_CVX_MATLAB(form, problem, eng, printout=printout)

    return problem


def call_and_output_CVX_MATLAB(form, problem, eng, printout=False):
    """Call and output the loadpath optimisation with CVX from MATLAB

    Parameters
    ----------
    form : ::class:: FormDiagram
        The form Diagram of the analysis
    problem : ::class:: Problem
        The Problem with relevant matrices and vectors
    eng : matlab.engine
        The matlab engine initiated to call the analysis
    printout : bool, optional
        Whether or not print results, by default False

    Returns
    -------
    obj : dict
        Dictionary with the returned values from the convex optimisation.
    """

    if len(problem.ind) < problem.m:
        print('Calling LP-Optimisation via CVX (MATLAB) with independents')
        fopt, qopt, exitflag, niter, status, sol_time = call_cvx_ind(problem, eng, printout=printout)
    else:
        print('Calling LP-Optimisation via CVX (MATLAB) with NO independents')
        fopt, qopt, exitflag, niter, status, sol_time = call_cvx(problem, eng, printout=printout)

    problem.q = qopt
    Xfinal = xyz_from_q(problem.q, problem.P[problem.free], problem.X[problem.fixed], problem.Ci, problem.Cit, problem.Cb)
    problem.X[problem.free, 2] = Xfinal[:, 2]

    # Update form diagram optimised

    i_uv = problem.i_uv

    i = 0
    for key in form.vertices():
        # form.vertex_attribute(key, 'x', problem.X[i, 0])
        # form.vertex_attribute(key, 'y', problem.X[i, 1])
        form.vertex_attribute(key, 'z', problem.X[i, 2])
        i = i + 1

    for c, qi in enumerate(list(problem.q.ravel())):
        u, v = i_uv[c]
        li = form.edge_length(u, v)
        form.edge_attribute((u, v), 'q', float(qi))
        form.edge_attribute((u, v), 'f', float(qi*li))

    form.attributes['loadpath'] = form.loadpath()
    compute_reactions(form)

    summary = True

    # Output dictionary

    output = {}
    output['fopt'] = fopt
    output['exitflag'] = exitflag
    output['status'] = status
    output['niter'] = niter
    output['sol_time'] = sol_time

    if printout or summary:
        print('\n' + '-' * 50)
        print('LOADPATH OPTIMISATION WITH CVX (MATLAB)')
        print('status    :', status)
        print('fopt (lp) : {0:.3f}'.format(fopt))
        print('n-iter    : {0}'.format(niter))
        print('q range   : {0:.3f} : {1:.3f}'.format(min(qopt)[0], max(qopt)[0]))
        print('sol. time : {0:.3f} sec'.format(sol_time))
        print('-' * 50 + '\n')

    return problem


def call_cvx(problem, eng, printout=False):
    """Call matlab with the matrices for the analysis

    Parameters
    ----------
    eng : matlab.engine
        The matlab engine initiated to call the analysis
    printout : bool, optional
        Whether or not print results, by default False

    Returns
    -------
    fopt : float
        Objective function value. Loadpath.
    qopt : array
        Force densities in the optimum.
    exitflag : int
        Whether or not optimisation worked.
    niter : int
        Number of iterations.
    status : str
        Message with statuss.
    sol_time : dict
        Time to solve optimisation.

    """

    # q, ind, dep, E, Edinv, Ei, C, Ct, Ci, Cit, Cf, U, V, p, px, py, pz, z, free, fixed, lh, sym, k, lb, ub, lb_ind, ub_ind, s, Wfree, x, y, qmax, i_uv, k_i, eng = args_cvx

    start_time1 = time.time()

    eng.workspace['m'] = problem.m

    # eng.workspace['C'] = matlab.double(problem.C.toarray().tolist())
    # eng.workspace['Ci'] = matlab.double(problem.Ci.toarray().tolist())
    # eng.workspace['Cb'] = matlab.double(problem.Cb.toarray().tolist())
    # eng.workspace['E'] = matlab.double(problem.E.tolist())

    # Passing as native arrays
    eng.workspace['C'] = problem.C.toarray()
    eng.workspace['Ci'] = problem.Ci.toarray()
    eng.workspace['Cb'] = problem.Cb.toarray()
    eng.workspace['E'] = problem.E

    eng.workspace['x'] = array(problem.X[:, 0].reshape(-1, 1))
    eng.workspace['y'] = array(problem.X[:, 1].reshape(-1, 1))

    eng.workspace['xb'] = problem.X[:, 0][problem.fixed].reshape(-1, 1)
    eng.workspace['yb'] = problem.X[:, 1][problem.fixed].reshape(-1, 1)

    eng.workspace['pz'] = problem.P[:, 2][problem.free].reshape(-1, 1)
    eng.workspace['p'] = problem.ph.reshape(-1, 1)

    eng.workspace['qmax'] = problem.qmax.reshape(-1, 1)
    eng.workspace['qmin'] = problem.qmin.reshape(-1, 1)

    # If intended to save the MATLAB engine
    # eng.save('/Users/mricardo/Documents/MATLAB/data.mat', nargout=0)

    start_time = time.time()

    if printout:
        eng.cvx_begin(nargout=0)
    else:
        eng.cvx_begin('quiet', nargout=0)
    # eng.cvx_begin(nargout=0)
    # eng.eval('cvx_begin', nargout=0)

    # Run the following lines in the terminal if you have trouble
    # future = matlab.engine.connect_matlab(background=True)
    # eng = future.result()
    # eng.cvx_begin(nargout=0)
    # print(matlab.engine.find_matlab())

    eng.variable('q(double(m))', nargout=0)
    # if objective == 'loadpath':
    eng.minimize('matrix_frac(pz, -(transpose(Ci)*diag(q)*Ci)) - transpose(x)*transpose(C)*diag(q)*Cb*xb - transpose(y)*transpose(C)*diag(q)*Cb*yb', nargout=0)
    # if objective == 'feasibility':
    # eng.minimize('1', nargout=0)
    eng.eval('q >= qmin', nargout=0)
    eng.eval('q <= qmax', nargout=0)
    eng.eval('E * q == p', nargout=0)
    eng.cvx_end(nargout=0)

    sol_time = time.time() - start_time

    fopt = eng.workspace['cvx_optval']
    qopt = array(eng.workspace['q'])
    status = eng.workspace['cvx_status']
    niter = eng.workspace['cvx_slvitr']

    elapsed_time = time.time() - start_time1
    print('Elapsed time on LP:', elapsed_time)

    if status != 'Infeasible':
        exitflag = 0
    else:
        exitflag = 1

    return fopt, qopt, exitflag, niter, status, sol_time


def call_cvx_ind(problem, eng, printout=True):
    """Call matlab considering independent edges.

    Parameters
    ----------
    eng : matlab.engine
        The matlab engine initiated to call the analysis
    printout : bool, optional
        Whether or not print results, by default False

    Returns
    -------
    fopt : float
        Objective function value. Loadpath.
    qopt : array
        Independent Force densities in the optimum.
    exitflag : int
        Whether or not optimisation worked.
    niter : int
        Number of iterations.
    status : str
        Message with statuss.
    sol_time : dict
        Time to solve optimisation.

    """

    # q, ind, dep, E, Edinv, Ei, C, Ct, Ci, Cit, Cf, U, V, p, px, py, pz, z, free, fixed, lh, sym, k, lb, ub, lb_ind, ub_ind, s, Wfree, x, y, qmax, i_uv, k_i, eng = args_cvx

    ind_ = [x+1 for x in problem.ind]
    dep_ = [x+1 for x in problem.dep]

    start_time = time.time()

    eng.workspace['C'] = matlab.double(problem.C.toarray().tolist())
    eng.workspace['Ci'] = matlab.double(problem.Ci.toarray().tolist())
    eng.workspace['Cb'] = matlab.double(problem.Cb.toarray().tolist())
    eng.workspace['Edinv'] = matlab.double(problem.Edinv.toarray().tolist())
    eng.workspace['Ei'] = matlab.double(problem.Ei.toarray().tolist())

    eng.workspace['m'] = problem.m

    eng.workspace['dep'] = matlab.double(dep_)
    eng.workspace['ind'] = matlab.double(ind_)
    eng.workspace['p'] = matlab.double(problem.ph.reshape(-1, 1).tolist())
    eng.workspace['xt'] = matlab.double(problem.X[:, 0].reshape(-1, 1).transpose().tolist())
    eng.workspace['yt'] = matlab.double(problem.X[:, 1].reshape(-1, 1).transpose().tolist())
    eng.workspace['xb'] = matlab.double(problem.X[:, 0][problem.fixed].reshape(-1, 1).tolist())
    eng.workspace['yb'] = matlab.double(problem.X[:, 1][problem.fixed].reshape(-1, 1).tolist())
    eng.workspace['pz'] = matlab.double(problem.P[:, 2][problem.free].reshape(-1, 1).tolist())
    eng.workspace['qmax'] = matlab.double(problem.qmax.reshape(-1, 1).tolist())
    eng.workspace['qmin'] = matlab.double(problem.qmin.reshape(-1, 1).tolist())

    sol_time0 = time.time()

    if printout:
        eng.cvx_begin(nargout=0)
    else:
        eng.cvx_begin('quiet', nargout=0)
    eng.variable('q(double(m))', nargout=0)
    eng.minimize('matrix_frac(pz, -transpose(Ci)*diag(q)*Ci) - xt*transpose(C)*diag(q)*Cb*xb - yt*transpose(C)*diag(q)*Cb*yb', nargout=0)  # with compression negative
    eng.eval('q >= qmin', nargout=0)
    eng.eval('q <= qmax', nargout=0)
    eng.eval('q(dep) == Edinv*(Ei*q(ind) - p)', nargout=0)
    eng.cvx_end(nargout=0)

    sol_time = time.time() - sol_time0

    fopt = eng.workspace['cvx_optval']
    status = eng.workspace['cvx_status']
    qopt = array(eng.workspace['q'])
    niter = int(eng.workspace['cvx_slvitr'])

    elapsed_time = time.time() - start_time
    print('Elapsed time on LP:', elapsed_time)

    # # If intended to save the MATLAB engine
    # eng.save('/Users/mricardo/Documents/MATLAB/data.mat', nargout=0)

    if status != 'Infeasible':
        exitflag = 0
    else:
        exitflag = 1

    return fopt, qopt, exitflag, niter, status, sol_time



from numpy import arange
from numpy import array
from numpy import linspace

from compas_tno.shapes import MeshDos
from compas_tno.shapes import rectangular_topology

import math


def domical_vault(xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.5, radius=None, center=None, tol=10e-6, t=0.0, discretisation=[100, 100]):
    """Set domical vault upper and lower meshes

    Parameters
    ----------
    xy_span : list, optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        [description], by default 0.5
    thk : float
        Thickness of the vault
    center : [float, float], optional
        Center of the dome in which the geometry of the vault is based, by default None in which the center point of the pattern is taken
    tol : float, optional
        Tolerance, by default 10e-6
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    discretisation : list|int, optional
        Level of discretisation of the shape, by default [100, 100]

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape
    """

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    dx = x1 - x0
    dy = y1 - y0
    d = math.sqrt(dx**2 + dy**2)
    r_diagonal = d/2

    if radius is None:
        radius = r_diagonal
    elif radius < r_diagonal:
        print('Check your radius!')
        raise Exception

    re = radius + thk/2
    ri = radius - thk/2

    if center is None:
        xc = (x1 + x0)/2
        yc = (y1 + y0)/2
    else:
        xc = center[0]
        yc = center[1]
        print('Center provided by the user:', center)

    density_x = discretisation[0]
    density_y = discretisation[1]
    x = arange(x0, x1 + dx/density_x, dx/density_x)
    y = arange(y0, y1 + dy/density_y, dy/density_y)

    index = 0
    uv_i = {}
    faces = []
    faces_i = []
    x1d = []
    y1d = []
    z1d = []
    zi1d = []
    ze1d = []

    for i in range(len(x)):
        for j in range(len(y)):
            uv_i[(i, j)] = index
            xi, yi = x[i], y[j]
            z2 = radius**2 - (xi - xc)**2 - (yi - yc)**2
            zi2 = ri**2 - (xi - xc)**2 - (yi - yc)**2
            ze2 = re**2 - (xi - xc)**2 - (yi - yc)**2
            z = math.sqrt(z2)
            ze = math.sqrt(ze2)
            if zi2 < 0:
                zi = -t
            else:
                zi = math.sqrt(zi2)

            x1d.append(xi)
            y1d.append(yi)
            z1d.append(z)
            zi1d.append(zi)
            ze1d.append(ze)

            if i < len(x) - 1 and j < len(y) - 1:
                p1 = (i, j)
                p2 = (i, j+1)
                p3 = (i+1, j)
                p4 = (i+1, j+1)
                face = [p1, p2, p4, p3]
                faces.append(face)
            index = index + 1

    for face in faces:
        face_i = []
        for uv in face:
            u, v = uv
            i = uv_i[(u, v)]
            face_i.append(i)
        faces_i.append(face_i)

    xyz = array([x1d, y1d, z1d]).transpose()
    middle = MeshDos.from_vertices_and_faces(xyz, faces_i)

    xyz = array([x1d, y1d, zi1d]).transpose()
    intrados = MeshDos.from_vertices_and_faces(xyz, faces_i)

    xyz = array([x1d, y1d, ze1d]).transpose()
    extrados = MeshDos.from_vertices_and_faces(xyz, faces_i)

    return intrados, extrados, middle


def parabolic_shell_highfields(xy_span=[[0.0, 10.0], [0.0, 10.0]], thk=0.5, hc=5.0, tol=10e-6, t=0.0, discretisation=[100, 100]):
    """ Set parabolic vault heights.

    Parameters
    ----------
    xy_span : list, optional
        [description], by default [[0.0, 10.0], [0.0, 10.0]]
    thk : float, optional
        [description], by default 0.5
    discretisation : list, optional
        [description], by default [10, 10]
    hc : float, optional
        Height in the middle point of the vault, by default 5.0
    tol : float, optional
        Tolerance, by default 10e-6
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    discretisation : list, optional
        [description], by default [100, 100]

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape

    Notes
    -------------
    Position of the quadrants is as in the schema below:

        Q3
    Q2      Q1
        Q4

    """

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]

    y1 = xy_span[1][1]
    y0 = xy_span[1][0]
    x1 = xy_span[0][1]
    x0 = xy_span[0][0]

    density_x = discretisation[0]
    density_y = discretisation[1]
    x = linspace(x0, x1, num=density_x+1, endpoint=True)  # arange(x0, x1 + dx/density_x, dx/density_x)
    y = linspace(y0, y1, num=density_y+1, endpoint=True)  # arange(y0, y1 + dy/density_y, dy/density_y)

    xi, yi, faces_i = rectangular_topology(x, y)

    zt = parabolic_shell_middle_update(xi, yi, t, xy_span=xy_span, hc=hc, tol=tol)
    xyzt = array([xi, yi, zt.flatten()]).transpose()
    middle = MeshDos.from_vertices_and_faces(xyzt, faces_i)

    zub, zlb = parabolic_shell_ub_lb_update(xi, yi, thk, t, xy_span=xy_span, hc=hc, tol=tol)
    xyzub = array([xi, yi, zub.flatten()]).transpose()
    xyzlb = array([xi, yi, zlb.flatten()]).transpose()
    extrados = MeshDos.from_vertices_and_faces(xyzub, faces_i)
    intrados = MeshDos.from_vertices_and_faces(xyzlb, faces_i)

    return intrados, extrados, middle


def parabolic_shell_middle_update(x, y, thk, xy_span=[[0.0, 10.0], [0.0, 10.0]], hc=5.0, tol=10e-6):
    """Update middle of a parabolic vault based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    hc : float, optional
        Height in the middle point of the vault, by default 8.0
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    zt : array
        Values of the middle surface in the points
    """
    print('WIP')
    zt = 0
    return zt


def parabolic_shell_ub_lb_update(x, y, thk, t, xy_span=[[0.0, 10.0], [0.0, 10.0]], hc=5.0, tol=10e-6):
    """Update upper and lower bounds of a parabolic vault based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    xy_span : [list[float], list[float]], optional
        xy-span of the shape, by default [[0.0, 10.0], [0.0, 10.0]]
    hc : float, optional
        Height in the middle point of the vault, by default 8.0
    tol : float, optional
        Tolerance, by default 10e-6

    Returns
    -------
    ub : array
        Values of the upper bound in the points
    lb : array
        Values of the lower bound in the points
    """
    print('WIP')
    ub = 0
    lb = 0
    return ub, lb



from compas_tno.algorithms import xyz_from_q
from compas_tno.algorithms import compute_reactions

from compas_tno.problems import initialise_problem_general
from compas_tno.problems import adapt_problem_to_fixed_diagram

import cvxpy as cp
from cvxpy import diag
from cvxpy import matrix_frac
from cvxpy import Minimize
from cvxpy import Problem


def run_optimisation_CVXPY(analysis):
    """ Run convex optimisation problem with CVXPY after going through the optimisation set up.

    Parameters
    ----------
    analysis : Analysis
        Analysis object with information about optimiser, form and shape.

    """

    form = analysis.form
    problem = analysis.optimiser.M
    find_inds = analysis.optimiser.settings.get('find_inds', False)
    printout = analysis.optimiser.settings.get('printout', False)

    problem = run_loadpath_from_form_CVXPY(form, problem=problem, find_inds=find_inds, printout=printout)

    return problem


def run_loadpath_from_form_CVXPY(form, problem=None, find_inds=False, printout=False):
    """ Run convex optimisation problem with CVXPY directly from the Form Diagram
    OBS: Requires installation of CVXPY and MOSEK

    Parameters
    ----------
    form : FormDiagram
        FormDiagram object with form containing full information.
    problem : Problem, optional
        The problem with matrices of interest, by default None
    find_inds : bool, optional
        Whether or not independents must be computed before the analysis, by default False
    printout : bool, optional
        Whether or not print results, by default False

    Returns
    -------
    obj : analysis
        Analysis object optimised.
    """

    if not problem:
        problem = initialise_problem_general(form)

    if find_inds:
        adapt_problem_to_fixed_diagram(problem, form)

    problem = call_and_output_CVXPY(form, problem, printout=printout)

    return problem


def call_and_output_CVXPY(form, problem, printout=False):
    """Call and output the loadpath optimisation with CVXPY

    Parameters
    ----------
    form : ::class:: FormDiagram
        The form Diagram of the analysis
    problem : ::class:: Problem
        The Problem with relevant matrices and vectors`
    printout : bool, optional
        Whether or not print results, by default False

    Returns
    -------
    output: dict
        Dictionary with results.
    """

    if len(problem.ind) < problem.m:
        print('Calling LP-Optimisation via CVXPY with independents')
        fopt, qopt, exitflag, niter, status, sol_time = call_cvxpy_ind(problem, printout=printout)
    else:
        print('Calling LP-Optimisation via CVXPY with NO independents')
        fopt, qopt, exitflag, niter, status, sol_time = call_cvxpy(problem, printout=printout)

    problem.q = qopt
    Xfinal = xyz_from_q(problem.q, problem.P[problem.free], problem.X[problem.fixed], problem.Ci, problem.Cit, problem.Cb)
    problem.X[problem.free, 2] = Xfinal[:, 2]

    # Update form diagram optimised

    i_uv = problem.i_uv

    i = 0
    for key in form.vertices():
        # form.vertex_attribute(key, 'x', problem.X[i, 0])
        # form.vertex_attribute(key, 'y', problem.X[i, 1])
        form.vertex_attribute(key, 'z', problem.X[i, 2])
        i = i + 1

    for c, qi in enumerate(list(problem.q.ravel())):
        u, v = i_uv[c]
        li = form.edge_length(u, v)
        form.edge_attribute((u, v), 'q', float(qi))
        form.edge_attribute((u, v), 'f', float(qi*li))

    form.attributes['loadpath'] = form.loadpath()
    compute_reactions(form)

    summary = True

    # Output dictionary

    output = {}
    output['fopt'] = fopt
    output['exitflag'] = exitflag
    output['status'] = status
    output['niter'] = niter
    output['sol_time'] = sol_time

    if printout or summary:
        print('\n' + '-' * 50)
        print('LOADPATH OPTIMISATION WITH CVXPY')
        print('status    :', status)
        print('fopt (lp) : {0:.3f}'.format(fopt))
        print('n-iter    : {0}'.format(niter))
        print('q range   : {0:.3f} : {1:.3f}'.format(min(qopt), max(qopt)))
        print('sol. time : {0:.3f} sec'.format(sol_time))
        print('-' * 50 + '\n')

    return problem


def call_cvxpy(problem, printout=False):
    """Call and output the loadpath optimisation with CVXPY

    Parameters
    ----------
    problem : ::class:: Problem
        The Problem with relevant matrices and vectors`
    printout : bool, optional
        Whether or not print results, by default False

    Returns
    -------
    fopt : float
        Objective function value. Loadpath.
    qopt : array
        Independent Force densities in the optimum.
    exitflag : int
        Whether or not optimisation worked.
    niter : int
        Number of iterations.
    status : str
        Message with statuss.
    sol_time : dict
        Time to solve optimisation.
    """

    # Retrieve important vector and matrices
    q = problem.q
    E = problem.E
    C = problem.C
    Ci = problem.Ci
    Cit = problem.Cit
    Cb = problem.Cb
    pz = problem.P[:, 2].reshape(-1, 1)
    ph = problem.ph
    free = problem.free
    fixed = problem.fixed
    qmin = problem.qmin
    qmax = problem.qmax
    x = problem.x0
    y = problem.y0
    m = problem.m

    # # If need to store the matrices/vectors in
    # dict_parameters = {}
    # dict_parameters['q'] = q.tolist()
    # dict_parameters['qmin'] = qmin.tolist()
    # dict_parameters['qmax'] = qmax.tolist()
    # dict_parameters['E'] = E.tolist()
    # dict_parameters['C'] = C.todense().tolist()
    # dict_parameters['Ci'] = Ci.todense().tolist()
    # dict_parameters['Cit'] = Cit.todense().tolist()
    # dict_parameters['Cb'] = Cb.todense().tolist()
    # dict_parameters['pz'] = pz.tolist()
    # dict_parameters['free'] = free
    # dict_parameters['fixed'] = fixed
    # dict_parameters['x'] = x.tolist()
    # dict_parameters['y'] = y.tolist()
    # dict_parameters['m'] = m

    # import json

    # with open('/Users/mricardo/compas_dev/compas_tno/data/data.json', 'w') as outfile:
    #     json.dump(dict_parameters, outfile)

    q = cp.Variable(m)
    # fobj = matrix_frac(pz[free], Cit@cp.diag(q)@Ci) + x.T@C.T@diag(q)@Cb@x[fixed] + y.T@C.T@diag(q)@Cb@y[fixed]  # for q positive
    fobj = matrix_frac(pz[free], - Cit@cp.diag(q)@Ci) - x.T@C.T@diag(q)@Cb@x[fixed] - y.T@C.T@diag(q)@Cb@y[fixed]  # for q negative
    objective = Minimize(fobj)

    horz = E@q == ph.flatten()
    pos = q >= qmin.flatten()
    maxq = q <= qmax.flatten()

    constraints = [horz, pos, maxq]

    prob = Problem(objective, constraints)
    prob.solve(solver='MOSEK', verbose=printout)
    # prob.solve(solver='MOSEK', verbose=True)

    # save output
    fopt = prob.value
    qopt = q.value
    status = prob.status
    niter = prob.solver_stats.num_iters
    sol_time = prob.solver_stats.solve_time

    if status not in ["infeasible", "unbounded"]:
        exitflag = 1
    else:
        exitflag = 0

    return fopt, qopt, exitflag, niter, status, sol_time


def call_cvxpy_ind(problem, printout=False):
    """Call and output the loadpath optimisation with CVXPY using independents

    Parameters
    ----------
    problem : ::class:: Problem
        The Problem with relevant matrices and vectors`
    printout : bool, optional
        Whether or not print results, by default False

    Returns
    -------
    fopt : float
        Objective function value. Loadpath.
    qopt : array
        Independent Force densities in the optimum.
    exitflag : int
        Whether or not optimisation worked.
    niter : int
        Number of iterations.
    status : str
        Message with statuss.
    sol_time : dict
        Time to solve optimisation.
    """

    # Retrieve important vector and matrices
    q = problem.q
    Edinv = problem.Edinv
    Ei = problem.Ei
    ph = problem.ph
    C = problem.C
    Ci = problem.Ci
    Cit = problem.Cit
    Cb = problem.Cb
    pz = problem.P[:, 2].reshape(-1, 1)
    free = problem.free
    fixed = problem.fixed
    dep = problem.dep
    ind = problem.ind
    qmin = problem.qmin
    qmax = problem.qmax
    x = problem.x0
    y = problem.y0
    m = problem.m

    q = cp.Variable(m)

    fobj = matrix_frac(pz[free], -Cit@cp.diag(q)@Ci) - x.T@C.T@diag(q)@Cb@x[fixed] - y.T@C.T@diag(q)@Cb@y[fixed]
    objective = Minimize(fobj)

    horz = q[dep] == Edinv@(Ei@q[ind] - ph.flatten())
    pos = q >= qmin.flatten()
    maxq = q <= qmax.flatten()

    constraints = [horz, pos, maxq]

    prob = Problem(objective, constraints)
    prob.solve(solver='MOSEK', verbose=printout)

    # save output
    fopt = prob.value
    qopt = q.value
    status = prob.status
    niter = prob.solver_stats.num_iters
    sol_time = prob.solver_stats.solve_time

    if status not in ["infeasible", "unbounded"]:
        exitflag = 1
    else:
        exitflag = 0

    return fopt, qopt, exitflag, niter, status, sol_time



from compas.datastructures import Datastructure

__all__ = ['Optimiser']


class Optimiser(Datastructure):

    """The ``Optimiser`` sets the parameters of the optimisation.

    Parameters
    ----------
    x0 : array
        The starting point of the optimisation.
    xopt : array
        The value of the variables that solves thhe optimisation problem.
    fopt : float
        The value of the objective function at the end of the optimisation process.
    message : str
        The message provided by the solver.
    M : :class:`~compas_tno.problems.Problem`
        The problem with the relevat matrices.
    time : float
        The time consumption for solving.
    niter : int
        The number of iterations.
    exitflag : int
        Whether or not the optimisation problem was solved successfully.
        Value ``0``represents a solved problem.
    log : str
        Log of the optimisation, available for some solves.

    Settings
    --------
    The main settings and default values are shown below:

    *  'library'           : ['Scipy', 'MATLAB', 'MMA', 'IPOPT', 'Scipy', ...]
    *  'solver'            : ['SLSQP', 'IPOPT', 'MMA', ...],
    *  'objective'         : ['min', 'max', 't', 'loadpath', 'bestfit', ...],
    *  'constraints'       : ['funicular', 'envelope', 'reac_bounds', ...],
    *  'variables'         : ['q', 'zb', 'xyb', 't', ...],
    *  'features'          : ['fixed', 'sym', ...],
    *  'find_inds'         : True,
    *  'exitflag'          : 1,
    *  'printout'          : True,
    *  'starting_point'    : 'current',
    *  'support_displacement': None,
    *  'gradient'          : True,
    *  'jacobian'          : True,
    *  'solver_options'    : {},
    *  'max_iter'          : 500,
    *  'qmin'              : -1e+4,
    *  'qmax'              : 1e-8,


    """

    def __init__(self):
        self.settings = {
            'solver': 'SLSQP',
            'objective': 'min',
            'constraints': ['funicular', 'envelope'],
            'features': ['fixed'],
            'variables': ['q', 'zb'],
            'find_inds': False,
            'printout': False,
            'plot': False,
            'starting_point': 'current',
            'solver_convex': 'CVXPY',
            'support_displacement': None,
            'gradient': True,
            'jacobian': True,
            'solver_options': {},
            'max_iter': 500,
            'qmin': -1e+4,
            'qmax': 1e-8,
        }
        self.clear_optimiser()

    @property
    def data(self):
        """dict : A data dict representing the shape data structure for serialization.
        """
        data = {
            'settings': self.settings,
            'x0': self.x0,
            'xopt': self.xopt,
            'fopt': self.fopt,
            'message': self.message,
            'niter': self.niter,
            'exitflag': self.exitflag,
            # 'log': self.log,
        }
        return data

    @data.setter
    def data(self, data):
        if 'data' in data:
            data = data['data']
        self.settings = data.get('settings') or {}

        self.x0 = data.get('x0', None)
        self.xopt = data.get('xopt', None)
        self.fopt = data.get('fopt', None)
        self.message = data.get('message', None)
        self.niter = data.get('niter', None)
        self.exitflag = data.get('exitflag', None)
        # self.log = data.get('log', None)

    @classmethod
    def create_minthk_optimiser(cls, solver='SLSQP', max_iter=500, printout=False, plot=False, starting_point='loadpath', derivatives=True):
        """Create a minimum thickness optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        solver : str, optional
            Which solver to use, by default 'SLSQP'. See Solvers page for more information.
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'
        derivatives : bool, optional
            Whether or not derivatives are computed by hand, by default True

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        optimiser = cls()
        optimiser.set_solver(solver)
        optimiser.set_constraints(['funicular', 'envelope'])
        optimiser.set_variables(['q', 'zb', 't'])
        optimiser.set_features(['fixed'])
        optimiser.set_objective('t')
        optimiser.set_display_options(plot=plot, printout=printout)
        optimiser.set_max_iterations(max_iter=max_iter)
        optimiser.set_gradient_options(gradient=derivatives, jacobian=derivatives)
        optimiser.set_starting_point(starting_point=starting_point)

        return optimiser

    @classmethod
    def create_minthrust_optimiser(cls, solver='SLSQP', max_iter=500, printout=False, plot=False, starting_point='loadpath'):
        """Create a minimum thickness optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        solver : str, optional
            Which solver to use, by default 'SLSQP'. See Solvers page for more information.
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        optimiser = cls()
        optimiser.set_solver(solver)
        optimiser.set_constraints(['funicular', 'envelope'])
        optimiser.set_variables(['q', 'zb'])
        optimiser.set_features(['fixed'])
        optimiser.set_objective('min')
        optimiser.set_display_options(plot=plot, printout=printout)
        optimiser.set_max_iterations(max_iter=max_iter)
        optimiser.set_gradient_options(gradient=True, jacobian=True)
        optimiser.set_starting_point(starting_point=starting_point)

        return optimiser

    @classmethod
    def create_maxhrust_optimiser(cls, solver='SLSQP', max_iter=500, printout=False, plot=False, starting_point='loadpath'):
        """Create a maximum thickness optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        solver : str, optional
            Which solver to use, by default 'SLSQP'. See Solvers page for more information.
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        optimiser = cls().create_minthrust_optimiser(solver=solver,
                                                     max_iter=max_iter,
                                                     printout=printout,
                                                     plot=plot,
                                                     starting_point=starting_point)
        optimiser.set_objective('max')

        return optimiser

    @classmethod
    def create_max_horload_optimiser(cls, solver='IPOPT', max_iter=500, printout=False, plot=False, starting_point='loadpath', max_lambd=1.0,
                                     load_direction=None, derivatives=True):
        """Create a minimum thickness optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        solver : str, optional
            Which solver to use, by default 'SLSQP'. See Solvers page for more information.
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'
        max_lambd : float, optional
            Maximum load multiplier, by default 1.0
        load_direction : array [2n x 1], optional
            Direction of the horizontal load, by default None
        derivatives : bool, optional
            Whether or not derivatives are computed by hand, by default True

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        optimiser = cls()
        optimiser.set_solver(solver)
        optimiser.set_constraints(['funicular', 'envelope'])
        optimiser.set_variables(['q', 'zb', 'lambdh'])
        optimiser.set_features(['fixed'])
        optimiser.set_objective('max_load')
        optimiser.set_display_options(plot=plot, printout=printout)
        optimiser.set_max_iterations(max_iter=max_iter)
        optimiser.set_gradient_options(gradient=derivatives, jacobian=derivatives)
        optimiser.set_starting_point(starting_point=starting_point)
        optimiser.set_additional_options(max_lambd=max_lambd, load_direction=load_direction)

        return optimiser

    @classmethod
    def create_bestfit_optimiser(cls, solver='SLSQP', max_iter=500, printout=False, plot=False, starting_point='loadpath', derivatives=True):
        """Create a bestfit optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        solver : str, optional
            Which solver to use, by default 'SLSQP'. See Solvers page for more information.
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        optimiser = cls()
        optimiser.set_solver(solver)
        optimiser.set_constraints(['funicular', 'envelope'])
        optimiser.set_variables(['q', 'zb'])
        optimiser.set_features(['fixed'])
        optimiser.set_objective('bestfit')
        optimiser.set_display_options(plot=plot, printout=printout)
        optimiser.set_max_iterations(max_iter=max_iter)
        optimiser.set_gradient_options(gradient=derivatives, jacobian=derivatives)
        optimiser.set_starting_point(starting_point=starting_point)

        return optimiser

    @classmethod
    def create_max_vertload_optimiser(cls, solver='IPOPT', max_iter=500, printout=False, plot=False, starting_point='loadpath', max_lambd=1.0,
                                      load_direction=None, derivatives=True):
        """Create a minimum thickness optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        solver : str, optional
            Which solver to use, by default 'SLSQP'. See Solvers page for more information.
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'
        max_lambd : float, optional
            Maximum load multiplier, by default 1.0
        load_direction : array [n x 1], optional
            Direction of the vertical external load, by default None
        derivatives : bool, optional
            Whether or not derivatives are computed by hand, by default True

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        optimiser = cls()
        optimiser.set_solver(solver)
        optimiser.set_constraints(['funicular', 'envelope'])
        optimiser.set_variables(['q', 'zb', 'lambdv'])
        optimiser.set_features(['fixed'])
        optimiser.set_objective('max_load')
        optimiser.set_display_options(plot=plot, printout=printout)
        optimiser.set_max_iterations(max_iter=max_iter)
        optimiser.set_gradient_options(gradient=derivatives, jacobian=derivatives)
        optimiser.set_starting_point(starting_point=starting_point)
        optimiser.set_additional_options(max_lambd=max_lambd, load_direction=load_direction)

        return optimiser

    @classmethod
    def create_lp_optimiser(cls, solver='MATLAB', printout=False, plot=False, max_iter=500):
        """Create a loadpath optimisation optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        form : FormDiagram
            _description_
        shape : Shape
            The shape cconstraining the problem
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        if solver not in ['MATLAB', 'CVXPY']:
            raise ValueError('For loadpath optimisation only MATLAB or CVXPY are possible solvers. See solvers page.')

        optimiser = cls()
        optimiser.set_solver(solver)
        optimiser.set_constraints(['funicular'])
        optimiser.set_variables(['q'])
        optimiser.set_features(['fixed'])
        optimiser.set_objective('loadpath')
        optimiser.set_display_options(plot=plot, printout=printout)
        optimiser.set_max_iterations(max_iter=max_iter)
        optimiser.set_gradient_options(gradient=True, jacobian=True)

        return optimiser

    @classmethod
    def create_compl_energy_optimiser(cls, solver='SLSQP', max_iter=500, printout=False, plot=False, starting_point='loadpath',
                                      support_displacement=None, Emethod='simplified'):
        """Create a linear complementary energy optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        solver : str, optional
            Which solver to use, by default 'SLSQP'. See Solvers page for more information.
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'
        support_displacement : array [nb x 3], optional
            Vector with the displacement applied to the supports, by default None
        Emethod : str, optional
            Whether or not internal deformations should be considered, by default 'simplified' which considers the material rigid

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        optimiser = cls()
        optimiser.set_solver(solver)
        optimiser.set_constraints(['funicular', 'envelope'])
        optimiser.set_variables(['q', 'zb'])
        optimiser.set_features(['fixed'])
        optimiser.set_objective('Ecomp-linear')
        optimiser.set_display_options(plot=plot, printout=printout)
        optimiser.set_max_iterations(max_iter=max_iter)
        optimiser.set_gradient_options(gradient=True, jacobian=True)
        optimiser.set_starting_point(starting_point=starting_point)
        optimiser.set_additional_options(support_displacement=support_displacement, Ecomp_method=Emethod)

        return optimiser

    @classmethod
    def create_quad_compl_energy_optimiser(cls, solver='SLSQP', max_iter=500, printout=False, plot=False,
                                           starting_point='loadpath', support_displacement=None, Emethod='simplified'):
        """Create a quadratic complementary energy optimiser to be sent with instructions to the Analysis.

        Parameters
        ----------
        solver : str, optional
            Which solver to use, by default 'SLSQP'. See Solvers page for more information.
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'
        support_displacement : array [nb x 3], optional
            Vector with the displacement applied to the supports, by default None
        Emethod : str, optional
            Whether or not internal deformations should be considered, by default 'simplified' which considers the material rigid

        Returns
        -------
        :class:`~compas_tno.optimisers.Optimiser`
            The Optimiser object

        """

        optimiser = cls()
        optimiser.set_solver(solver)
        optimiser.set_constraints(['funicular', 'envelope'])
        optimiser.set_variables(['q', 'zb'])
        optimiser.set_features(['fixed'])
        optimiser.set_objective('Ecomp-nonlinear')
        optimiser.set_display_options(plot=plot, printout=printout)
        optimiser.set_max_iterations(max_iter=max_iter)
        optimiser.set_gradient_options(gradient=True, jacobian=True)
        optimiser.set_starting_point(starting_point=starting_point)
        optimiser.set_additional_options(support_displacement=support_displacement, Ecomp_method=Emethod)

        return optimiser

    def set_objective(self, objective='min'):
        """Set the objective function for the problem.

        Parameters
        ----------
        objective : str, optional
            Name of the objective function to call, by default 'min'
            For a list of possible objective functions see [link].

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(objective, str):
            raise ValueError('Please provide the name of the objective')

        self.settings['objective'] = objective

    def set_constraints(self, constraints=[]):
        """Set the constraints for the problem.

        Parameters
        ----------
        constraints : [str], optional
            List wihth the names of the constraints to activate, by default []
            For a list of possible constraints functions see [link].

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(constraints, list):
            raise ValueError('Please provide a list with the name of constraints')

        self.settings['constraints'] = constraints

    def set_variables(self, variables=[]):
        """Set the variables for the problem.

        Parameters
        ----------
        variables : [str], optional
            List wihth the names of the variables to activate, by default []
            For a list of possible variables functions see [link].

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(variables, list):
            raise ValueError('Please provide a list with the name of variables')

        self.settings['variables'] = variables

    def set_features(self, features=[]):
        """Set the features for the problem.

        Parameters
        ----------
        features : [str], optional
            List with the names of the features to activate, by default []
            For a list of possible features functions see [link].

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(features, list):
            raise ValueError('Please provide a list with the name of features')

        self.settings['features'] = features

    def set_solver(self, solver='SLSQP'):
        """Set the features for the problem.

        Parameters
        ----------
        solver : str, optional
            Set the solver to use.
            For a list of possible solvers see [link].

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(solver, str):
            raise ValueError('Please provide the name of the solver')

        self.settings['solver'] = solver

    def set_max_iterations(self, max_iter=500):
        """Set the features for the problem.

        Parameters
        ----------
        max_iter : int, optional
            Number of iterations, by default 500.

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(max_iter, int):
            raise ValueError('Please provide a int as max iteration')

        self.settings['max_iter'] = max_iter

    def set_starting_point(self, starting_point='loadpath'):
        """Set the starating point of the problem.

        Parameters
        ----------
        starting_point : str, optional
            Set the starting point to use.
            For a list of possible starting points see [link].

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(starting_point, str):
            raise ValueError('Please provide the name of the starting point')

        self.settings['starting_point'] = starting_point

    def set_display_options(self, plot=False, printout=True):
        """Set the display options of the optimisation

        Parameters
        ----------
        plot : bool, optional
            If plots with partial states should be evokes, by default False
        printout : bool, optional
            If prints with iterations should appear in the screen, by default True

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(plot, bool):
            raise ValueError('Please provide a bool')
        if not isinstance(printout, bool):
            raise ValueError('Please provide a bool')

        self.settings['plot'] = plot
        self.settings['printout'] = printout

    def set_gradient_options(self, gradient=True, jacobian=True):
        """Set the display options of the optimisation

        Parameters
        ----------
        gradient : bool, optional
            If gradient should be provided analytically, by default True
        jacobian : bool, optional
            If jacobian should be provided analytically, by default True

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(gradient, bool):
            raise ValueError('Please provide a bool')
        if not isinstance(jacobian, bool):
            raise ValueError('Please provide a bool')

        self.settings['gradient'] = gradient
        self.settings['jacobian'] = jacobian

    def set_axis_symmetry(self, axis_sym=[]):
        """Set the axis of symmetry.

        Parameters
        ----------
        axis_sym : list, optional
            List with the lines of symmetry

        Returns
        -------
        None
            Optimiser is modified in place
        """

        if not isinstance(axis_sym, list):
            raise ValueError('Please provide a list with the name of features')

        self.settings['axis_sym'] = axis_sym

    def set_additional_options(self, **kwargs):
        """Set the additional options of the optimisation.
        """

        if kwargs:
            for key in kwargs:
                self.settings[key] = kwargs[key]

    def clear_optimiser(self):
        """Initiate an empty Optimiser objects, or clear Previous results stored in the Optimiser object.
        Necessary to perform sequential optimisation with the same analysis object
        """

        self.x0 = None
        self.xopt = None
        self.fopt = None
        self.M = None
        self.message = None
        self.time = None
        self.niter = None
        self.exitflag = None
        self.callback = None

        return

    def __str__(self):
        tpl = "-"*20 + '\nOptimiser with settings:\n'
        for key in self.settings:
            tpl = tpl + key + ' : ' + str(self.settings[key]) + '\n'
        tpl = tpl + "-"*20 + '\n'
        # tpl = "<Optimiser with settings {}>".format(self.settings)
        return tpl



from numpy import zeros
from numpy import identity
from numpy import hstack
from numpy import vstack

from scipy.sparse.linalg import splu
from scipy.sparse import diags

from compas_tno.problems.bounds_update import dub_dlb_update
from compas_tno.problems.bounds_update import db_update

from compas_tno.algorithms import q_from_variables
from compas_tno.algorithms import xyz_from_q


def d_fconstr(fconstr, x0, eps, *args):
    """Jacobian matrix approximated using finite differences.

    Parameters
    ----------
    fconstr : callable
        Function with the constraints
    x0 : array
        Point to compute the pertubation
    eps : float
        Size of the pertubations

    Returns
    -------
    dfdx : array
        Jacobian of the constraints computed using finite diferences
    """

    fval = fconstr(x0, *args).reshape(-1, 1)
    m = len(fval)
    n = len(x0)
    dfdx = zeros((m, n))
    for i in range(m):
        for j in range(n):
            diff = zeros((n, 1))
            diff[j] = eps
            dfdx[i, j] = (fconstr(x0 + diff, *args).reshape(-1, 1) - fval)[i]/diff[j]
    return dfdx


def sensitivities_wrapper(variables, M):
    """Jacobian matrix computed analytically based on the constraints and variables assigned.

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    [type]
        [description]

    Notes
    -----
        Observe the order in which variables are added.
    """

    if isinstance(M, list):
        M = M[0]

    # variables
    k = M.k  # number of force variables
    n = M.n  # number of vertices
    m = M.m
    nb = len(M.fixed)  # number of fixed vertices
    nbz = 0
    nbxy = 0
    delta = 0.0
    thk = M.thk  # Introduce this because it might be necessary
    t = M.shape.datashape['t']
    lambdh = 1.0

    qid = variables[:k].reshape(-1, 1)
    check = k

    if 'xyb' in M.variables:
        xyb = variables[check:check + 2*nb]
        check = check + 2*nb
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
        nbxy = nb
    if 'zb' in M.variables:
        zb = variables[check: check + nb]
        check = check + nb
        M.X[M.fixed, [2]] = zb.flatten()
        nbz = nb
    if 't' in M.variables or 'n' in M.variables:
        thk = variables[check: check + 1]
        check = check + 1
    if 'lambdh' in M.variables:
        lambdh = variables[check: check + 1]
        M.P[:, [0]] = lambdh * M.px0
        M.P[:, [1]] = lambdh * M.py0
        M.d = lambdh * M.d0
        # M.d[M.dep] = -M.Edinv.dot(vstack([M.P[M.free_x, 0].reshape(-1, 1), M.P[M.free_y, 1].reshape(-1, 1)]))  # make this line shorter
        check = check + 1
    if 'lambdv' in M.variables:
        lambdv = variables[check: check + 1]
        M.P[:, [2]] = lambdv * M.pzv + M.pz0
        check = check + 1
    if 'tub' in M.variables:
        tub = variables[check: check + n]
        M.tub = tub
        check = check + n
    if 'tlb' in M.variables:
        tlb = variables[check: check + n]
        M.tlb = tlb
        check = check + n
    if 'tub_reac' in M.variables:
        tub_reac = variables[check: check + 2*nb].reshape(-1, 1)
        M.tub_reac = tub_reac
        check = check + 2*nb
    if 'delta' in M.variables:
        delta = float(variables[check: check + 1])
        check = check + 1

    M.q = q_from_variables(qid, M.B, M.d)

    # update geometry
    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    q = M.q

    M.U = diags(M.C.dot(M.X[:, 0]))  # U = diag(Cx)
    M.V = diags(M.C.dot(M.X[:, 1]))  # V = diag(Cy)
    M.W = diags(M.C.dot(M.X[:, 2]))  # W = diag(Cz)

    # initialize jac matrix
    deriv = zeros([0, M.k])

    Q = diags(q.ravel())
    CitQCi = M.Cit.dot(Q).dot(M.Ci)
    SPLU_D = splu(CitQCi)
    nlin_fun = 0
    nlin_limitxy = 0
    nlin_env = 0
    nlin_reacbounds = 0
    dqdqi = M.B

    A = zeros((n, nb))
    CitQCf = M.Cit.dot(Q).dot(M.Cb).toarray()
    A[M.free] = SPLU_D.solve(-CitQCf)
    A[M.fixed] = identity(nb)

    dxdqi = zeros((n, k))  # TODO: rename this to dxdqii to be more consistent
    dydqi = zeros((n, k))
    dzdqi = zeros((n, k))
    db_column = zeros((nlin_reacbounds, 1))

    # ------------ Adding rows to the jacobian matrix based on constraints activated ------------

    if 'funicular' in M.constraints:
        deriv = vstack([deriv, dqdqi, - dqdqi])
        nlin_fun = 2 * m

    if 'envelopexy' in M.constraints:
        # jacobian of in constraints on x
        dxidq = SPLU_D.solve(-M.Cit.dot(M.U).toarray())
        dxidqi = dxidq.dot(dqdqi)
        dxdqi[M.free] = dxidqi
        deriv = vstack([deriv, dxdqi, - dxdqi])

        # jacobian of in constraints on y
        dyidq = SPLU_D.solve(-M.Cit.dot(M.V).toarray())
        dyidqi = dyidq.dot(dqdqi)
        dydqi[M.free] = dyidqi
        deriv = vstack([deriv, dydqi, - dydqi])

        nlin_limitxy = 4 * n

    if 'envelope' in M.constraints:
        # jacobian of in constraints on z
        dzidq = SPLU_D.solve(-M.Cit.dot(M.W).toarray())
        dzidqi = dzidq.dot(dqdqi)
        dzdqi[M.free] = dzidqi

        if 'update-envelope' in M.features:
            dzmaxdt, dzmindt, dzmaxdx, dzmindx, dzmaxdy, dzmindy = dub_dlb_update(M.X[:, 0], M.X[:, 1], thk, t, M.shape, None, None, M.s, M.variables)
            dzmaxdq = dzmaxdx.dot(dxdqi) + dzmaxdy.dot(dydqi)
            dzmindq = dzmindx.dot(dxdqi) + dzmindy.dot(dydqi)
            deriv = vstack([deriv, dzdqi - dzmindq, dzmaxdq - dzdqi])
        else:
            deriv = vstack([deriv, dzdqi, - dzdqi])

        nlin_env = 2 * n

    if 'reac_bounds' in M.constraints:
        CbQC = M.Cb.transpose().dot(Q).dot(M.C)

        dRxdq = M.Cb.transpose().dot(M.U).dot(dqdqi) + CbQC.dot(dxdqi)
        dRydq = M.Cb.transpose().dot(M.V).dot(dqdqi) + CbQC.dot(dydqi)
        dRzdq = M.Cb.transpose().dot(M.W).dot(dqdqi) + CbQC.dot(dzdqi)

        dRzdzb = CbQC.dot(A)

        R = CbQC.dot(M.X) - M.P[M.fixed]

        dslope_dind = zeros((2 * nb, len(M.ind)))
        dslope_dzb = zeros((2 * nb, nb))
        # dslope_dlambd = zeros((2 * nb, 1))

        for i in range(len(M.fixed)):
            i_ = nb + i
            zbi = M.X[M.fixed, 2][i]
            # px0i = M.P[M.fixed, 0][i]
            # py0i = M.P[M.fixed, 1][i]

            signe_x = 1.0
            signe_y = 1.0
            signe_z = 1.0
            if R[i, 0] < 0:
                signe_x = -1.0
            if R[i, 1] < 0:
                signe_y = -1.0
            if R[i, 2] < 0:
                signe_z = -1.0

            dslope_dzb[i, i] = - 1 * abs(R[i, 0]/R[i, 2])
            dslope_dzb[i] += signe_z * zbi * abs(R[i, 0])/R[i, 2]**2 * dRzdzb[:, i]

            # dslope_dlambd[i] = - zbi / abs(R[i, 2]) * signe_x * (- px0i)

            dslope_dind[i] = zbi * signe_x * (-R[i, 2] * dRxdq[i] + R[i, 0] * dRzdq[i]) / R[i, 2]**2 / signe_z

            dslope_dzb[i_, i] = - 1 * abs(R[i, 1]/R[i, 2])
            dslope_dzb[i_] += signe_z * zbi * abs(R[i, 1])/R[i, 2]**2 * dRzdzb[:, i]

            # dslope_dlambd[i_] = - zbi / abs(R[i, 2]) * signe_y * (- py0i)

            dslope_dind[i_] = zbi * signe_y * (-R[i, 2] * dRydq[i] + R[i, 1] * dRzdq[i]) / R[i, 2]**2 / signe_z

        deriv = vstack([deriv, dslope_dind])

        if 't' in M.variables or 'n' in M.variables:
            db = db_update(M.x0, M.y0, thk, M.fixed, M.shape, M.b, M.variables)
            db_column = vstack([db[:, 0].reshape(-1, 1), db[:, 1].reshape(-1, 1)])

        nlin_reacbounds = 2 * nb

    if 'displ_map' in M.constraints:

        if delta:
            dhdq = M.E + delta * M.Ed
        else:
            dhdq = M.E
        deriv = vstack([deriv, dhdq, -dhdq])

        nlin_displ_map = 2 * dhdq.shape[0]

    # ------------ Adding columns to the jacobian matrix based on variables activated activated ------------
    # ------------ Note: length of the column to be defined through the "marking parameters" ---------------

    if nbxy or nbz:  # add a column to the derivatives to count the variables zb or xyb
        Anull = zeros((n, nb))
        if nbxy:
            deriv = hstack([deriv, vstack([zeros((nlin_fun, nb)), A, -A, Anull, -Anull, Anull, -Anull, zeros((nlin_reacbounds, nb))])])
            deriv = hstack([deriv, vstack([zeros((nlin_fun, nb)), Anull, -Anull, A, -A, Anull, -Anull, zeros((nlin_reacbounds, nb))])])
        if nbz:
            addcolumn = zeros((nlin_fun + nlin_limitxy, nb))
            if 'envelope' in M.constraints:
                addcolumn = vstack([addcolumn, A, -A])
            if 'reac_bounds' in M.constraints:
                addcolumn = vstack([addcolumn, dslope_dzb])
            if 'displ_map' in M.constraints:
                addcolumn = vstack([addcolumn, zeros((nlin_displ_map, nb))])
            deriv = hstack([deriv, addcolumn])

    if 't' in M.variables or 'n' in M.variables:  # add a column to the derivatives to count the variable t (thickness)
        if 'update-envelope' in M.features:
            pass
        else:
            dzmaxdt, dzmindt = dub_dlb_update(M.x0, M.y0, thk, t, M.shape, M.ub0, M.lb0, M.s, M.variables)[:2]

        dXdt = vstack([zeros((nlin_fun + nlin_limitxy, 1)), -dzmindt, +dzmaxdt, db_column])
        deriv = hstack([deriv, dXdt])

        if 'displ_map' in M.constraints:
            raise NotImplementedError()

    if 'lambdh' in M.variables:  # add a column to the derivatives to count the variable lambd (hor-multiplier)

        dqdlambd = zeros((M.m, 1))
        dxdlambd = zeros((n, 1))
        dydlambd = zeros((n, 1))
        dzdlambd = zeros((n, 1))
        if 'fixed' in M.features:
            dqdlambd = M.d0
            dzdq = zeros((n, m))
            dzdq[M.free] = dzidq
            dzdlambd = dzdq.dot(dqdlambd)
            dXdlambd = vstack([dqdlambd, -dqdlambd, dzdlambd, -dzdlambd])
            # dXdlambd = vstack([dqdlambd, -dqdlambd, zeros((nlin_limitxy, 1)), -dzdlambd, +dzdlambd])
        else:
            dxdlambd[M.free] = SPLU_D.solve(M.px0[M.free]).reshape(-1, 1)
            dydlambd[M.free] = SPLU_D.solve(M.py0[M.free]).reshape(-1, 1)

            if 'update-envelope' in M.features:
                dzmaxdlambd = dzmaxdx.dot(dxdlambd) + dzmaxdy.dot(dydlambd)
                dzmindlambd = dzmindx.dot(dxdlambd) + dzmindy.dot(dydlambd)
                dXdlambd = vstack([zeros((nlin_fun, 1)), dxdlambd, - dxdlambd, dydlambd, - dydlambd, - dzmindlambd, +dzmaxdlambd])
            else:
                dXdlambd = vstack([zeros((nlin_fun, 1)), dxdlambd, - dxdlambd, dydlambd, - dydlambd, zeros((nlin_env, 1))])

        if 'reac_bounds' in M.constraints:

            dRxdlambd = M.Cb.transpose().dot(M.U).dot(dqdlambd) + CbQC.dot(dxdlambd) - M.px0[M.fixed]
            dRydlambd = M.Cb.transpose().dot(M.V).dot(dqdlambd) + CbQC.dot(dydlambd) - M.py0[M.fixed]
            dRzdlambd = M.Cb.transpose().dot(M.W).dot(dqdlambd) + CbQC.dot(dzdlambd)
            dslope_dlambdh = zeros((2 * nb, 1))
            for i in range(nb):
                i_ = nb + i
                zbi = M.X[M.fixed, 2][i]
                signe_x = 1.0
                signe_y = 1.0
                signe_z = 1.0
                if R[i, 0] < 0:
                    signe_x = -1.0
                if R[i, 1] < 0:
                    signe_y = -1.0
                if R[i, 2] < 0:
                    signe_z = -1.0

                dslope_dlambdh[i] = zbi * signe_x * (-R[i, 2] * dRxdlambd[i] + R[i, 0] * dRzdlambd[i]) / R[i, 2]**2 / signe_z
                dslope_dlambdh[i_] = zbi * signe_y * (-R[i, 2] * dRydlambd[i] + R[i, 1] * dRzdlambd[i]) / R[i, 2]**2 / signe_z

            dXdlambd = vstack([dXdlambd, dslope_dlambdh])

        if 'displ_map' in M.constraints:
            raise NotImplementedError()

        deriv = hstack([deriv, dXdlambd])

    if 'lambdv' in M.variables:  # add a column to the derivatives to count the variable lambdv (vertical load multiplier)

        dzdlambdv = zeros((n, 1))
        dzdlambdv[M.free] = SPLU_D.solve(M.pzv[M.free]).reshape(-1, 1)

        if 'fixed' in M.features:
            dXdlambd = vstack([zeros((nlin_fun, 1)), dzdlambdv, - dzdlambdv])
        else:
            if 'update-envelope' in M.features:
                dzmaxdlambd = dzmaxdx.dot(dxdlambd) + dzmaxdy.dot(dydlambd)
                dzmindlambd = dzmindx.dot(dxdlambd) + dzmindy.dot(dydlambd)
                dXdlambd = vstack([zeros((nlin_fun, 1)), dxdlambd, - dxdlambd, dydlambd, - dydlambd, - dzmindlambd, +dzmaxdlambd])
            else:
                dXdlambd = vstack([zeros((nlin_fun, 1)), zeros((nlin_limitxy, 1)), dzdlambdv, - dzdlambdv])

        if 'reac_bounds' in M.constraints:
            dRzdlambdv = CbQC.dot(dzdlambdv) - M.pzv[M.fixed]
            dslope_dlambdv = zeros((2 * nb, 1))
            for i in range(nb):
                i_ = nb + i
                zbi = M.X[M.fixed, 2][i]
                signe_z = 1.0
                if R[i, 2] < 0:
                    signe_z = -1.0

                dslope_dlambdv[i] = signe_z * zbi * abs(R[i, 0])/R[i, 2]**2 * dRzdlambdv[i]
                dslope_dlambdv[i_] = signe_z * zbi * abs(R[i, 1])/R[i, 2]**2 * dRzdlambdv[i]
            dXdlambd = vstack([dXdlambd, dslope_dlambdv])

        if 'displ_map' in M.constraints:
            dXdlambd = vstack([dXdlambd, zeros((nlin_displ_map, 1))])

        deriv = hstack([deriv, dXdlambd])

    if 'tub' in M.variables:  # add a column to the derivatives to count the variable tub (max_section)

        nconst = deriv.shape[0]
        dXdtub = zeros((nconst, n))
        startline = nlin_fun + nlin_limitxy
        endline = startline + nlin_env

        In = identity(n)
        I0 = 0.0 * In
        Mt = vstack([I0, In])

        dXdtub[startline: endline, :] = Mt

        deriv = hstack([deriv, dXdtub])

        if 'displ_map' in M.constraints:
            raise NotImplementedError()

    if 'tlb' in M.variables:  # add a column to the derivatives to count the variable tub (max_section)

        nconst = deriv.shape[0]
        dXdtub = zeros((nconst, n))
        startline = nlin_fun + nlin_limitxy
        endline = startline + nlin_env

        In = identity(n)
        I0 = 0.0 * In
        Mt = vstack([In, I0])

        dXdtub[startline: endline, :] = Mt

        deriv = hstack([deriv, dXdtub])

        if 'displ_map' in M.constraints:
            raise NotImplementedError()

    if 'tub_reac' in M.variables:  # add a column to the derivatives to count the variable tub (max_section)

        nconst = deriv.shape[0]
        dXdtreacub = zeros((nconst, 2*nb))
        startline = nlin_fun + nlin_limitxy + nlin_env
        endline = startline + nlin_reacbounds

        Inb = identity(nb)  # check if the modulus need to be added here
        Inb0 = 0.0 * Inb
        Mt = hstack([vstack([Inb, Inb0]), vstack([Inb0, Inb])])

        dXdtreacub[startline: endline, :] = Mt

        deriv = hstack([deriv, dXdtreacub])

        if 'displ_map' in M.constraints:
            raise NotImplementedError()

    if 'delta' in M.variables:

        nconst = deriv.shape[0]
        addcolumn = zeros((nconst, 1))
        startline = nlin_fun + nlin_limitxy + nlin_env + nlin_reacbounds
        endline = startline + nlin_displ_map

        dhddelta = M.Ed @ M.q
        dhddelta = vstack([dhddelta, - dhddelta])
        addcolumn[startline: endline] = dhddelta

        deriv = hstack([deriv, addcolumn])

    return deriv



from numpy import vstack
from numpy import multiply
from numpy import divide
from numpy import zeros
from scipy.sparse import diags

from compas_tno.problems.bounds_update import ub_lb_update
from compas_tno.problems.bounds_update import b_update

from compas_tno.algorithms import q_from_variables
from compas_tno.algorithms import xyz_from_q
from compas_tno.algorithms import weights_from_xyz


def constr_wrapper(variables, M):
    """Wrapper of the constraints assigned.

    Parameters
    ----------
    variables : array [(]k x 1]
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    constraints: array [m x 1]
        Vector with the value of the m constraints in the point.
    """

    if isinstance(M, list):
        M = M[0]

    k = M.k
    n = M.n
    nb = M.nb
    check = k
    qid = variables[:k].reshape(-1, 1)
    lambdh = 1.0
    delta = 0.0
    thk = M.thk
    t = M.shape.datashape['t']

    if 'xyb' in M.variables:
        xyb = variables[check:check + 2*nb]
        check = check + 2*nb
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = variables[check: check + nb]
        check = check + nb
        M.X[M.fixed, [2]] = zb.flatten()
    if 't' in M.variables or 'n' in M.variables:
        thk = variables[check: check + 1]
        check = check + 1
    if 'lambdh' in M.variables:
        lambdh = variables[check: check + 1]
        M.P[:, [0]] = lambdh * M.px0
        M.P[:, [1]] = lambdh * M.py0
        M.d = lambdh * M.d0
        # M.d[M.dep] = -M.Edinv.dot(vstack([M.P[M.free_x, 0].reshape(-1, 1), M.P[M.free_y, 1].reshape(-1, 1)]))  # make this line shorter
        check = check + 1
    if 'lambdv' in M.variables:
        lambdv = variables[check: check + 1]
        M.P[:, [2]] = lambdv * M.pzv + M.pz0
        check = check + 1
    if 'tub' in M.variables:
        tub = variables[check: check + n].reshape(-1, 1)
        M.tub = tub
        check = check + n
    if 'tlb' in M.variables:
        tlb = variables[check: check + n].reshape(-1, 1)
        M.tlb = tlb
        check = check + n
    if 'tub_reac' in M.variables:
        tub_reac = variables[check: check + 2*nb].reshape(-1, 1)
        M.tub_reac = tub_reac
        check = check + 2*nb
    if 'delta' in M.variables:
        delta = float(variables[check: check + 1])
        check = check + 1

    M.q = q_from_variables(qid, M.B, M.d)

    if 'update-loads' in M.features:
        M.P[:, 2] = -1 * weights_from_xyz(M.X, M.F, M.V0, M.V1, M.V2, thk=M.thk, density=M.ro)

    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    # if 'fixed' in M.features:
    #     M.X[M.free, 2] = X_free[:, 2]
    # else:
    #     M.X[M.free] = X_free

    constraints = zeros([0, 1])

    if 'funicular' in M.constraints:
        qmin = M.q.reshape(-1, 1) - M.qmin
        qmax = M.qmax - M.q.reshape(-1, 1)
        constraints = vstack([constraints, qmin, qmax])

    if 'envelopexy' in M.constraints:
        # constraints in x
        xmin = (M.X[:, 0] - M.xlimits[:, 0]).reshape(-1, 1)
        xmax = (M.xlimits[:, 1] - M.X[:, 0]).reshape(-1, 1)
        constraints = vstack([constraints, xmin, xmax])

        # constraints in y
        ymin = (M.X[:, 1] - M.ylimits[:, 0]).reshape(-1, 1)
        ymax = (M.ylimits[:, 1] - M.X[:, 1]).reshape(-1, 1)
        constraints = vstack([constraints, ymin, ymax])

    if 'envelope' in M.constraints:
        # constraints in z
        if 'update-envelope' in M.features:
            M.ub, M.lb = ub_lb_update(M.X[:, 0], M.X[:, 1], thk, t, M.shape, None, None, M.s, M.variables)
        elif 't' in M.variables or 'n' in M.variables:
            M.ub, M.lb = ub_lb_update(M.x0, M.y0, thk, t, M.shape, M.ub0, M.lb0, M.s, M.variables)
        else:
            pass
        zmin = (M.X[:, 2] - M.lb.flatten()).reshape(-1, 1)
        zmax = (M.ub.flatten() - M.X[:, 2]).reshape(-1, 1)
        if 'tub' in M.variables:
            zmax = zmax + tub
        if 'tlb' in M.variables:
            zmin = zmin + tlb
        constraints = vstack([constraints, zmin, zmax])

    if 'reac_bounds' in M.constraints:
        # constraints in compute_reactions
        if 't' in M.variables:
            M.b = b_update(M.x0, M.y0, thk, M.fixed, M.shape, M.b, M.variables)
        elif 'n' in M.variables:
            raise NotImplementedError
        else:
            pass

        CbQC = M.Cb.transpose().dot(diags(M.q.flatten())).dot(M.C)
        R = CbQC.dot(M.X) - M.P[M.fixed]
        Rx = abs(M.b[:, [0]].reshape(-1, 1)) - multiply(M.X[:, [2]][M.fixed] - M.s[M.fixed], abs(divide(R[:, [0]], R[:, [2]]).reshape(-1, 1)))  # >= 0
        Ry = abs(M.b[:, [1]].reshape(-1, 1)) - multiply(M.X[:, [2]][M.fixed] - M.s[M.fixed], abs(divide(R[:, [1]], R[:, [2]]).reshape(-1, 1)))  # >= 0

        if 'tub_reac' in M.variables:
            Rx = Rx + abs(tub_reac[:nb])
            Ry = Ry + abs(tub_reac[nb:])

        constraints = vstack([constraints, Rx, Ry])

    if 'displ_map' in M.constraints:
        if delta:
            pos = (M.E @ M.q + delta * M.Ed @ M.q - M.ph).reshape(-1, 1)
        else:
            pos = (M.E @ M.q - M.ph).reshape(-1, 1)
        # pos = (M.E @ M.q - M.ph).reshape(-1, 1)
        neg = -pos

        constraints = vstack([constraints, pos, neg])

    return constraints.flatten()



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc


__commandname__ = "TNO__restart"


def RunCommand(is_interactive):

    scene = sc.sticky['TNO']['scene']
    if not scene:
        return

    proxy = sc.sticky['TNO']['proxy']
    if not proxy:
        return

    scene.purge()
    proxy.stop_server()
    proxy.start_server()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)
